/*
 ** xkaapi
 ** 
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** francois.broquedis@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "libgomp.h"

typedef struct komp_trampoline_arg_t {
  void (*fn) (void *);
  void *data;
  komp_dataenv_icv_t* parent_icvs;
} komp_trampoline_arg_t;

static void komp_trampoline_fn( void* a )
{
  komp_trampoline_arg_t* arg = (komp_trampoline_arg_t*)a;

#if KOMP_USE_ICV
  kaapi_processor_t* proc = kaapi_self_processor();
  kaapi_context_t* context = proc->context;
  komp_dataenv_icv_t* newicvs = komp_newdataenv_icvs( proc, arg->parent_icvs );
  komp_update_implicit_icvs(newicvs);
  context->komp_icvs = newicvs;
#endif

  arg->fn(arg->data);
  
#if KOMP_USE_ICV
  context->komp_icvs = 0;
#endif
}


/* TODO: declare new ICVS if num_threads == 1 ? 
*/
void
GOMP_parallel_start (
    void (*fn) (void *), 
    void *data, 
    unsigned num_threads
)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
#if KOMP_USE_ICV
  komp_dataenv_icv_t* env = komp_get_dataenv_icv( kproc );
  num_threads = komp_get_thread_parallel_region( num_threads, 1, env );
  komp_proc_bind_t procbind = env->h_bind;
#else
  if (num_threads == 0) num_threads = kaapi_get_concurrency();
  komp_proc_bind_t procbind = komp_proc_bind_default;
#endif
  /* what does means true here ? convert to default that has mapping to kaapi */
  if (procbind == komp_proc_bind_true)
    procbind = komp_proc_bind_default;

  kaapi_team_t* team = kaapi_begin_parallel(
      num_threads,
      (kaapi_procbind_t)procbind
  );

  komp_trampoline_arg_t* arg = (komp_trampoline_arg_t*)&team->ws.data[0];
  arg->fn = fn;
  arg->data = data;
#if KOMP_USE_ICV
  arg->parent_icvs = env;
#endif

  /* start parallel team, except the current context */
  kaapi_start_parallel( team, komp_trampoline_fn, arg, 0);

#if KOMP_USE_ICV
  /* set the ICV to the newly created context */
  kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;
  context->komp_icvs = komp_newdataenv_icvs( kproc, env );
  komp_update_implicit_icvs(context->komp_icvs);
#endif
}


void 
GOMP_parallel_end (void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;

#if 0
{
  static int counter = 0;
  char filename[128];
  snprintf(filename,128,"/tmp/%s_gompparallel_%i.dot",getenv("USER"), ++counter);
  FILE* file = fopen(filename,"w");
  kaapi_assert_debug(file !=0);
  kaapi_context_print_dot( file, context);
  fclose(file);
}
#endif
  if (context->ws.count_for_end != context->ws.count_for)
    komp_loop_workfini( (komp_thread_t*)kaapi_context2thread(context) );
  kaapi_assert_debug(context->ws.count_for_end == context->ws.count_for);

  kaapi_end_parallel( kproc->team );
}


/* GOMP 4.0 */
void
GOMP_parallel (void (*fn) (void *), void *data, unsigned num_threads, unsigned int flags)
{
#if KOMP_USE_ICV
  kaapi_processor_t* kproc = kaapi_self_processor();
  komp_dataenv_icv_t* env = komp_get_dataenv_icv( kproc );
  komp_proc_bind_t procbind = env->h_bind;
  komp_proc_bind_t bind_var = flags & 7;
  if ((procbind != komp_proc_bind_false) && (bind_var != komp_proc_bind_false))
    env->h_bind = bind_var;
#endif
  GOMP_parallel_start(fn, data, num_threads);
  fn (data);
  GOMP_parallel_end();
}
