/*
 ** xkaapi
 ** 
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "libgomp.h"
#include <string.h>

//#define USE_WORKLOAD 1


/* This version pass to each created task a pointer to the parent context.
   It is correct because:
   - the parent thread will always waits for completion of the child tasks,
   - a context has a scope at least equal to the task that creates it.
*/
typedef struct GOMP_trampoline_task_arg {
  void                     (*fn) (void *);
  void*                     data;
#if KOMP_USE_ICV
  komp_dataenv_icv_t*       icvs;
#endif
  int                       n_depend;
  komp_access_t*            depends;
} GOMP_trampoline_task_arg;

static void GOMP_trampoline_task(
  struct kaapi_task_t* task, kaapi_thread_t* th
)
{
  GOMP_trampoline_task_arg* taskarg = (GOMP_trampoline_task_arg*)kaapi_task_getargs(task);
#if KOMP_USE_ICV
  kaapi_processor_t*  kproc  = kaapi_self_processor();
  kaapi_context_t*    thread = kproc->context;
  komp_dataenv_icv_t* save_icvs = thread->komp_icvs;
  thread->komp_icvs = taskarg->icvs;
#endif
#if 0
  printf("Begin execute task \n");
  int i;
  for (i=0; i<taskarg->n_depend; ++i)
  {
    printf("[%i] mode=%i, ", i, taskarg->depends[i].reserved2 );
  }
  printf("\n");
#endif

  taskarg->fn(taskarg->data);

#if 0
  printf("End execute task \n");
#endif

#if KOMP_USE_ICV
  thread->komp_icvs = save_icvs;
#endif
}


/*
*/
//static int counter = 0;
void GOMP_task(
  void (*fn) (void *),
  void *data, 
  void (*cpyfn) (void *, void *),
  long arg_size, 
  long arg_align, 
  bool if_clause,
  unsigned flags __attribute__((unused)),
  void **depend /* new GOMP-4.0 */
)
{
//printf("%i GOMP_task\n", ++counter);
  size_t count_depend;
  size_t clo_size;
  kaapi_processor_t* kproc = komp_self_processor();
  kaapi_context_t* context = kproc->context;
  kaapi_task_flag_t kaapi_flag = KAAPI_TASK_FLAG_DEFAULT;

  /* if if_clause if false, but it exists dependencies, then 
     sync previous task and execute the task as a function call.
     If orphan, execute the task as a function.
  */
  if (!if_clause || (kproc->team ==0))
  {
#if KOMP_USE_ICV
    komp_dataenv_icv_t* save_icvs = komp_get_dataenv_icv( kproc );
    context->komp_icvs = komp_newdataenv_icvs( kproc, save_icvs );
#endif
    /* if dependencies = sync previous task */
    if (flags & 8)
      kaapi_sched_sync( kaapi_context2thread(context) );

    if (!cpyfn)
      fn (data);
    else
    {
      char buf[arg_size + arg_align - 1];
      char *arg = (char *) (((uintptr_t) buf + (uintptr_t)arg_align - 1)
                            & ~(uintptr_t) (arg_align - 1));
      cpyfn (arg, data);
      fn (arg);
    }
#if KOMP_USE_ICV
    context->komp_icvs = save_icvs;
#endif
    return;
  }

  if (flags & 8) /* depend clauses */
  {
    /* see libgomp, task.c  */
    count_depend = (uintptr_t) depend[0];
    clo_size = arg_size + sizeof(komp_access_t)*count_depend;
    /* depend[i+2] is the address on which to compute dependency */
  }
  else
  {
    clo_size = arg_size;
    count_depend = 0;
    kaapi_flag = KAAPI_TASK_FLAG_INDPENDENT;
  }

  void* userarg = kaapi_data_push_align( kaapi_context2thread(context),
                    (uint32_t)clo_size, (uintptr_t)arg_align);
  if (cpyfn)
    cpyfn(userarg, data);
  else
    memcpy(userarg, data, arg_size);

  GOMP_trampoline_task_arg* arg = kaapi_data_push(kaapi_context2thread(context), sizeof(GOMP_trampoline_task_arg));
  arg->fn             = fn;
  arg->data           = userarg;
#if KOMP_USE_ICV
  komp_dataenv_icv_t* env = komp_get_dataenv_icv( kproc );
  arg->icvs               = komp_newdataenv_icvs( kproc, env );
#endif
  arg->n_depend = 0;
  if (count_depend !=0)
  {
    size_t i;
    size_t count_out = (uintptr_t) depend[1];
    arg->n_depend = (int)count_depend;
    arg->depends  = (komp_access_t*)(arg_size + ((char*)userarg));
    for (i=0; i<count_depend; ++i)
    {
      komp_access_mode_t mode;
      if (i < count_out)
        mode = KOMP_ACCESS_MODE_RW; /* it is a READ WRITE in the sens of Kaapi */
      else
        mode = KOMP_ACCESS_MODE_R;
      arg->depends[i].data      = depend[i+2];
      arg->depends[i].version   = 0;
      arg->depends[i].reserved1 = 0; /* owner task */
      arg->depends[i].reserved2 = mode; /* mode */
      arg->depends[i].reserved3 = 0; /* next */
    }
  }

  kaapi_thread_t* thread = kaapi_context2thread(context);
  kaapi_task_t* task = thread->sp;
  if (!_kaapi_has_enough_taskspace(thread,1))
    task = kaapi_thread_slow_pushtask(thread);
  task->body        = (kaapi_task_body_t)GOMP_trampoline_task;
  task->state.steal = (kaapi_task_body_t)GOMP_trampoline_task;
  task->arg         = arg;
  task->u.dummy     = 0;
  task->fmt         = 0;
  task->next        = 0;
  task->prev        = 0;
  KAAPI_ATOMIC_WRITE( &task->wc, 0);
#if KOMP_USE_ICV
  komp_dataenv_icv_t* icvs = komp_get_dataenv_icv( kaapi_thread2context(thread)->proc );
  if (icvs->prio)
  {
    task->u.s.priority = icvs->prio;
    kaapi_flag |= KAAPI_TASK_FLAG_PRIORITY;
    icvs->prio = 0;
  }
  if (icvs->preferred_ld)
  {
    task->u.s.site     = (uint8_t)icvs->preferred_ld;
    icvs->preferred_ld = 0;
  }
  if (icvs->flag & KOMP_FLAG_LLC_AWARE)
    kaapi_flag |= KAAPI_TASK_FLAG_LLC_AWARE;
#endif
  task->u.s.flag = kaapi_flag;
  kaapi_writemem_barrier();
  ++thread->sp;
}

void GOMP_taskwait (void)
{
  kaapi_thread_t* thread = kaapi_self_thread();
  if (thread)
    kaapi_sched_sync(thread);
}

/* GOMP4.0 */
void
GOMP_taskgroup_start (void)
{
  komp_taskgroup_begin();
}


void
GOMP_taskgroup_end (void)
{
  komp_taskgroup_end();
}


/* format definition for any GOMP task */
static size_t
kaapi_gomp_task_format_get_size(const struct kaapi_format_t* fmt, const void* sp)
{
  return sizeof(GOMP_trampoline_task_arg);
}

static void
kaapi_gomp_task_format_task_copy(const struct kaapi_format_t* fmt, void* sp_dest, const void* sp_src)
{
  memcpy( sp_dest, sp_src, sizeof(GOMP_trampoline_task_arg) );
}

static
unsigned int kaapi_gomp_task_format_get_count_params(const struct kaapi_format_t* fmt, const void* sp)
{
  GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  return arg->n_depend;
}

static
kaapi_access_mode_t kaapi_gomp_task_format_get_mode_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  return arg->depends[i].reserved2;
}

static
void* kaapi_gomp_task_format_get_data_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  return arg->depends[i].data;
}

static
kaapi_access_t* kaapi_gomp_task_format_get_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  return (kaapi_access_t*)&arg->depends[i];
}

static
void kaapi_gomp_task_format_set_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_access_t* a
)
{
  GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  arg->depends[i] = *(komp_access_t*)a;
}

static
const struct kaapi_format_t* kaapi_gomp_task_format_get_fmt_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  return (kaapi_format_t*)komp_char_format;
}

static
void kaapi_gomp_task_format_get_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp, kaapi_memory_view_t* view
)
{
  //GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
  kaapi_memory_view_make1d( view, 0, 1, sizeof(char) );
}

static
void kaapi_gomp_task_format_set_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_memory_view_t* view
)
{
  //GOMP_trampoline_task_arg* arg = (GOMP_trampoline_task_arg*)sp;
}

static
void kaapi_gomp_task_format_reducor(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const void* v
)
{
  abort();
}

static
void kaapi_gomp_task_format_redinit(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp, void* v
)
{
  abort();
}

static kaapi_adaptivetask_splitter_t	kaapi_gomp_task_fnc_get_splitter(
  const struct kaapi_format_t* fmt, const void* sp
)
{
  return 0;
}


static kaapi_format_t* kaapi_gomp_task_format = 0;

__attribute__((constructor))
static void kaapi_gomp_task_format_constructor(void)
{
  if (kaapi_gomp_task_format !=0) return;
  kaapi_gomp_task_format = kaapi_format_allocate();
  kaapi_format_taskregister_func(
                                 kaapi_gomp_task_format,
                                 GOMP_trampoline_task,
                                 "kaapi_gomp_task_format",
                                 0,
                                 kaapi_gomp_task_format_get_size,
                                 kaapi_gomp_task_format_task_copy,
                                 kaapi_gomp_task_format_get_count_params,
                                 kaapi_gomp_task_format_get_mode_param,
                                 kaapi_gomp_task_format_get_data_param,
                                 kaapi_gomp_task_format_get_access_param,
                                 kaapi_gomp_task_format_set_access_param,
                                 kaapi_gomp_task_format_get_fmt_param,
                                 kaapi_gomp_task_format_get_view_param,
                                 kaapi_gomp_task_format_set_view_param,
                                 kaapi_gomp_task_format_reducor,
                                 kaapi_gomp_task_format_redinit,
                                 kaapi_gomp_task_fnc_get_splitter
                                 );
}

