/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012, 2013, 2014 INRIA.
**
** Contributors :
**
** francois.broquedis@imag.fr
** thierry.gautier@inrialpes.fr
**
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <libiomp.h>


/* Always returns true. (see kmp_csupport.c:173) */
kmp_int32 __kmpc_ok_to_fork(ident_t *loc)
{
  TRACE_ENTER();
  TRACE_LEAVE();

  return 1;
}

/* Returns the id of the current thread. */
kmp_int32 __kmpc_global_thread_num(ident_t *loc)
{
  return 0; /* should return either pointer to kprocessor, but cannot be cast without
        guarantee to kmp_int32. Return kid to the team structure ? So not use
     */
}
#if 0
{
  kmp_int32 retval;
  TRACE_ENTER();
  kaapi_processor_t* kproc = kaapi_self_processor();
  if (kproc == 0)
    retval = -1;
  else
    retval = (kmp_int32)kproc->kid;
  TRACE_LEAVE();
  return retval;
}
#endif

#if !KOMP_USE_ICV
 __thread kmp_int32 tls_num_threads = 0;
#endif

void __kmpc_push_num_threads (ident_t *loc, kmp_int32 global_tid, kmp_int32 num_threads)
{
  TRACE_ENTER();
#if KOMP_USE_ICV
  komp_set_num_threads( num_threads );
#else
  tls_num_threads = num_threads;
#endif
  TRACE_LEAVE();
}

struct microtask_arg {
  microtask_t microtask;
  kmp_int32 argc;
  void** argv;
  komp_dataenv_icv_t* icvs;
};


static void __kmpc_wrapper_parallel_kaapi( void* _a )
{
  struct microtask_arg* arg= (struct microtask_arg*)_a;
  int gtid = kaapi_self_kid(); 

#if KOMP_USE_ICV
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;
  komp_dataenv_icv_t* newicvs = komp_newdataenv_icvs( kproc, arg->icvs );
  komp_update_implicit_icvs(newicvs);
  context->komp_icvs = newicvs;
#endif

  __kmp_invoke_microtask( arg->microtask, gtid, gtid, (int)arg->argc, (void**)arg->argv );

#if KOMP_USE_ICV
  context->komp_icvs = 0;
#endif
}

void __kmpc_fork_call (ident_t *loc, kmp_int32 argc, kmpc_micro microtask, ...) 
{
  TRACE_ENTER();

#if KOMP_USE_ICV
  kmp_int32 num_threads = komp_get_max_threads();
#else
  kmp_int32 num_threads = tls_num_threads;
  if (num_threads ==0) num_threads = komp_get_max_threads();
#endif

//printf("[iomp] num thread:%i, max thread:%i\n", num_threads, omp_get_max_threads());

  struct microtask_arg arg;
  void  **argv;
  va_list ap;
  int i;


  /* file args for the microtask_arg */
  arg.microtask = (microtask_t)microtask;
  arg.argv = alloca( sizeof(void*)*argc);
  arg.argc = argc;
  arg.icvs = komp_self_icvs();
  argv = arg.argv;
  va_start(ap, microtask);

  for( i=argc-1; i >= 0; --i )
    /* TODO: revert workaround for Intel(R) 64 tracker #96 */
#if (KMP_ARCH_X86_64 || KMP_ARCH_ARM) && KMP_OS_LINUX
    *argv++ = va_arg( *ap, void * );
#else
    *argv++ = va_arg( ap, void * );
#endif
  va_end(ap);

  komp_fork_call(num_threads, komp_proc_bind_default, __kmpc_wrapper_parallel_kaapi, &arg);

#if !KOMP_USE_ICV
  /* reset num_threads clause for next region */
  tls_num_threads = 0;
#endif

  TRACE_LEAVE();
}


void __kmpc_serialized_parallel(ident_t *loc, kmp_int32 global_tid)
{
  /* TODO: Est-ce qu'on doit créer une team dans ce cas-là? (accès
   * aux ICVs etc) */
  TRACE_ENTER();
  fprintf (stderr, "%s is deprecated!\n", __FUNCTION__);
  exit (-1);
  TRACE_LEAVE();
}

void __kmpc_end_serialized_parallel(ident_t *loc, kmp_int32 global_tid)
{
  TRACE_ENTER();
  fprintf (stderr, "%s is deprecated!\n", __FUNCTION__);
  exit (-1);
  TRACE_LEAVE();
}

kmp_int32 __kmpc_single(ident_t *loc, kmp_int32 global_tid)
{
  TRACE_ENTER();
  TRACE_LEAVE();
  return kaapi_team_single(kaapi_self_team(), 0);
}

void __kmpc_end_single(ident_t *loc, kmp_int32 global_tid)
{
  TRACE_ENTER();
  TRACE_LEAVE();
}

kmp_int32 __kmpc_master(ident_t *loc, kmp_int32 global_tid)
{
  TRACE_ENTER();
  TRACE_LEAVE();
  return kaapi_team_master(kaapi_self_team(), 0);
}

void __kmpc_end_master(ident_t *loc, kmp_int32 global_tid)
{
  TRACE_ENTER();
  TRACE_LEAVE();
}
