/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** francois.broquedis@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef _KAAPI_LIBGOMP_
#define _KAAPI_LIBGOMP_

#define KAAPI_KOMP_TRACE 0
#include "komp_lock-internal.h"

#include "libkomp_impl.h"

#ifdef HAVE_VISIBILITY_HIDDEN
# pragma GCC visibility push(hidden)
#endif

/* going back to the previous visibility, ie "default" */
#ifdef HAVE_VISIBILITY_HIDDEN
# pragma GCC visibility pop
#endif


#include "libgomp_g.h"

enum gomp_cancel_kind
{
  GOMP_CANCEL_PARALLEL = 1,
  GOMP_CANCEL_LOOP = 2,
  GOMP_CANCEL_FOR = GOMP_CANCEL_LOOP,
  GOMP_CANCEL_DO = GOMP_CANCEL_LOOP,
  GOMP_CANCEL_SECTIONS = 4,
  GOMP_CANCEL_TASKGROUP = 8
};


#ifdef HAVE_VERSION_SYMBOL
extern void komp_init_lock_30 (struct komp_lock_t *) __KOMP_NOTHROW;
extern void komp_destroy_lock_30 (struct komp_lock_t *) __KOMP_NOTHROW;
extern void komp_set_lock_30 (struct komp_lock_t *) __KOMP_NOTHROW;
extern void komp_unset_lock_30 (struct komp_lock_t *) __KOMP_NOTHROW;
extern int  komp_test_lock_30 (struct komp_lock_t *) __KOMP_NOTHROW;
extern void komp_init_nest_lock_30 (struct komp_nest_lock_t *) __KOMP_NOTHROW;
extern void komp_destroy_nest_lock_30 (struct komp_nest_lock_t *) __KOMP_NOTHROW;
extern void komp_set_nest_lock_30 (struct komp_nest_lock_t *) __KOMP_NOTHROW;
extern void komp_unset_nest_lock_30 (struct komp_nest_lock_t *) __KOMP_NOTHROW;
extern int  komp_test_nest_lock_30 (struct komp_nest_lock_t *) __KOMP_NOTHROW;

#if 0 /* not yet implemented */
extern void komp_init_lock_25 (struct komp_lock_25_t *) __KOMP_NOTHROW;
extern void komp_destroy_lock_25 (struct komp_lock_25_t *) __KOMP_NOTHROW;
extern void komp_set_lock_25 (struct komp_lock_25_t *) __KOMP_NOTHROW;
extern void komp_unset_lock_25 (struct komp_lock_25_t *) __KOMP_NOTHROW;
extern int  komp_test_lock_25 (struct komp_lock_25_t *) __KOMP_NOTHROW;
extern void komp_init_nest_lock_25 (struct komp_nest_lock_25_t *) __KOMP_NOTHROW;
extern void komp_destroy_nest_lock_25 (struct komp_nest_lock_25_t *) __KOMP_NOTHROW;
extern void komp_set_nest_lock_25 (struct komp_nest_lock_25_t *) __KOMP_NOTHROW;
extern void komp_unset_nest_lock_25 (struct komp_nest_lock_25_t *) __KOMP_NOTHROW;
extern int  komp_test_nest_lock_25 (struct komp_nest_lock_25_t *) __KOMP_NOTHROW;
#endif

# define komp_lock_symver30(fn) \
  __asm (".symver k" #fn "_30, " #fn "@@OMP_3.0");
# define komp_lock_symver(fn) \
  komp_lock_symver30(fn) \
  __asm (".symver k" #fn "_25, " #fn "@OMP_1.0");

#else
# define komp_init_lock_30 omp_init_lock
# define komp_destroy_lock_30 omp_destroy_lock
# define komp_set_lock_30 omp_set_lock
# define komp_unset_lock_30 omp_unset_lock
# define komp_test_lock_30 omp_test_lock
# define komp_init_nest_lock_30 omp_init_nest_lock
# define komp_destroy_nest_lock_30 omp_destroy_nest_lock
# define komp_set_nest_lock_30 omp_set_nest_lock
# define komp_unset_nest_lock_30 omp_unset_nest_lock
# define komp_test_nest_lock_30 omp_test_nest_lock

#define komp_lock_symver30(fn)
#define komp_lock_symver(fn)

#endif



#endif // #ifndef _KAAPI_LIBGOMP_
