/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012, 2013, 2014 INRIA.
**
** Contributors :
**
** francois.broquedis@imag.fr
** thierry.gautier@inrialpes.fr
**
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef LIBIOMP_H
#define LIBIOMP_H

#include <stdio.h>
#include <stdint.h>

#include "libkomp_impl.h"

#define kmp_int32 int32_t
#define kmp_uint32 uint32_t
#define kmp_int64 int64_t
#define kmp_uint64 uint64_t

#ifdef TRACE_LIB
#define TRACE_ENTER(...) fprintf(stderr, "[debug] {%s} Entering %s.\n", __FILE__, __FUNCTION__);
#define TRACE_LEAVE(...) fprintf(stderr, "[debug] {%s} Leaving %s.\n", __FILE__, __FUNCTION__);
#else
#define TRACE_ENTER(...)
#define TRACE_LEAVE(...)
#endif


typedef struct ident {
    kmp_int32 reserved_1;
    kmp_int32 flags;
    kmp_int32 reserved_2;
    kmp_int32 reserved_3;
    char *psource;
} ident_t;

typedef kmp_int32 (* kmp_routine_entry_t)( kmp_int32, void * );

typedef struct kmp_task_t {                   /* GEH: Shouldn't this be aligned somehow? */
    void *              shareds;            /**< pointer to block of pointers to shared vars   */
    kmp_routine_entry_t routine;            /**< pointer to routine to call for executing task */
    kmp_int32           part_id;            /**< part id for the task                          */
#if 1//OMP_40_ENABLED
    kmp_routine_entry_t destructors;        /* pointer to function to invoke deconstructors of firstprivate C++ objects */
#endif // OMP_40_ENABLED
    /*  private vars  */
} kmp_task_t;


typedef struct kmp_taskdata_t {
#if KOMP_USE_ICV
    komp_dataenv_icv_t* icvs;
#endif
    uintptr_t           size;
} kmp_taskdata_t;


enum sched_type {
    kmp_sch_lower                     = 32,   /**< lower bound for unordered values */
    kmp_sch_static_chunked            = 33,
    kmp_sch_static                    = 34,   /**< static unspecialized */
    kmp_sch_dynamic_chunked           = 35,
    kmp_sch_guided_chunked            = 36,   /**< guided unspecialized */
    kmp_sch_runtime                   = 37,
    kmp_sch_auto                      = 38,   /**< auto */
    kmp_sch_trapezoidal               = 39,

    /* accessible only through KMP_SCHEDULE environment variable */
    kmp_sch_static_greedy             = 40,
    kmp_sch_static_balanced           = 41,
    /* accessible only through KMP_SCHEDULE environment variable */
    kmp_sch_guided_iterative_chunked  = 42,
    kmp_sch_guided_analytical_chunked = 43,

    kmp_sch_steal                     = 44,   /**< accessible only through KMP_SCHEDULE environment variable */

    /* accessible only through KMP_SCHEDULE environment variable */
    kmp_sch_upper                     = 45,   /**< upper bound for unordered values */

    kmp_sch_adaptive                  = 46,

    kmp_ord_lower                     = 64,   /**< lower bound for ordered values, must be power of 2 */
    kmp_ord_static_chunked            = 65,
    kmp_ord_static                    = 66,   /**< ordered static unspecialized */
    kmp_ord_dynamic_chunked           = 67,
    kmp_ord_guided_chunked            = 68,
    kmp_ord_runtime                   = 69,
    kmp_ord_auto                      = 70,   /**< ordered auto */
    kmp_ord_trapezoidal               = 71,
    kmp_ord_upper                     = 72,   /**< upper bound for ordered values */

#if OMP_40_ENABLED
    /* Schedules for Distribute construct */
    kmp_distribute_static_chunked     = 91,   /**< distribute static chunked */
    kmp_distribute_static             = 92,   /**< distribute static unspecialized */
#endif

    /*
     * For the "nomerge" versions, kmp_dispatch_next*() will always return
     * a single iteration/chunk, even if the loop is serialized.  For the
     * schedule types listed above, the entire iteration vector is returned
     * if the loop is serialized.  This doesn't work for gcc/gcomp sections.
     */
    kmp_nm_lower                      = 160,  /**< lower bound for nomerge values */

    kmp_nm_static_chunked             = (kmp_sch_static_chunked - kmp_sch_lower + kmp_nm_lower),
    kmp_nm_static                     = 162,  /**< static unspecialized */
    kmp_nm_dynamic_chunked            = 163,
    kmp_nm_guided_chunked             = 164,  /**< guided unspecialized */
    kmp_nm_runtime                    = 165,
    kmp_nm_auto                       = 166,  /**< auto */
    kmp_nm_trapezoidal                = 167,

    /* accessible only through KMP_SCHEDULE environment variable */
    kmp_nm_static_greedy              = 168,
    kmp_nm_static_balanced            = 169,
    /* accessible only through KMP_SCHEDULE environment variable */
    kmp_nm_guided_iterative_chunked   = 170,
    kmp_nm_guided_analytical_chunked  = 171,
    kmp_nm_static_steal               = 172,  /* accessible only through OMP_SCHEDULE environment variable */

    kmp_nm_ord_static_chunked         = 193,
    kmp_nm_ord_static                 = 194,  /**< ordered static unspecialized */
    kmp_nm_ord_dynamic_chunked        = 195,
    kmp_nm_ord_guided_chunked         = 196,
    kmp_nm_ord_runtime                = 197,
    kmp_nm_ord_auto                   = 198,  /**< auto */
    kmp_nm_ord_trapezoidal            = 199,
    kmp_nm_upper                      = 200,  /**< upper bound for nomerge values */

    kmp_sch_default = kmp_sch_static  /**< default scheduling algorithm */
};

static inline void print_ident(const ident_t *loc) {
    fprintf(stderr, "ident_t *loc: %p\n", loc);
    fprintf(stderr, "\tloc->reserved_1 = %i\n", loc->reserved_1);
    fprintf(stderr, "\tloc->flags \t= %i\n", loc->flags);
    fprintf(stderr, "\tloc->reserved_2 = %i\n", loc->reserved_2);
    fprintf(stderr, "\tloc->reserved_3 = %i\n", loc->reserved_3);
    fprintf(stderr, "\tloc->psource \t= %s\n", loc->psource);
}

/* init.c */
extern void __kmpc_begin(ident_t *loc, kmp_int32 flags);
extern void __kmpc_end(ident_t *loc);

/* parallel.c */
typedef void (*kmpc_micro)(kmp_int32 *global_tid, kmp_int32 *bound_tid, ...);
typedef void (*microtask_t)( int *gtid, int *npr, ... );
extern int __kmp_invoke_microtask( microtask_t pkfn, int gtid, int npr, int argc, void *argv[] );

kmp_int32 __kmpc_ok_to_fork(ident_t *loc);
extern void __kmpc_push_num_threads(ident_t *loc, kmp_int32 global_tid, kmp_int32 num_threads);
extern void __kmpc_fork_call(ident_t *loc, kmp_int32 argc, kmpc_micro microtask, ...);
kmp_int32 __kmpc_global_thread_num(ident_t *loc);
kmp_int32 __kmpc_single(ident_t *loc, kmp_int32 global_tid);
void __kmpc_end_single(ident_t *loc, kmp_int32 global_tid);
kmp_int32 __kmpc_master(ident_t *loc, kmp_int32 global_tid);
void __kmpc_end_master(ident_t *loc, kmp_int32 global_tid);

/* barrier.c */
extern void __kmpc_barrier(ident_t *loc, kmp_int32 global_tid);

/* task.c */
extern kmp_task_t*
__kmpc_omp_task_alloc( ident_t *loc_ref, kmp_int32 gtid, kmp_int32 flags,
                       size_t sizeof_kmp_task_t, size_t sizeof_shareds,
                       kmp_routine_entry_t task_entry );
extern kmp_int32
__kmpc_omp_task( ident_t *loc_ref, kmp_int32 gtid, kmp_task_t * new_task );
extern kmp_int32
__kmpc_omp_taskwait( ident_t *loc_ref, kmp_int32 gtid );

/* see Intel omp.h for the definition of the types : */
typedef struct komp_lock_t
{
  unsigned char _x[64] 
    __attribute__((__aligned__(8)));
} komp_lock_t;

typedef komp_lock_t omp_lock_t;

typedef struct komp_nest_lock_t
{
  unsigned char _x[80] 
    __attribute__((__aligned__(8)));
} komp_nest_lock_t;

typedef komp_nest_lock_t omp_nest_lock_t;

#endif /* LIBIOMP_H */

