/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com / fabien.lementec@imag.fr
** vincent.danjean@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef _KAAPI_IMPL_H
#define _KAAPI_IMPL_H 1


/* to use new aggregation protocol */
#define KAAPI_USE_FLAT_CC_SYNC 1

#if defined(KAAPI_DEBUG)
#define KAAPI_RETURN_ERROR(err,val) \
  {\
    if (err != val) { printf("%s %i: error\n", __FILE__, __LINE__); }\
    return err; \
  }
#else
#define KAAPI_RETURN_ERROR(err,val) \
  {\
    return err; \
  }
#endif


/* Mark that we compile source of the library.
   Only used to avoid to include public definitition of some types.
*/
#define KAAPI_COMPILE_SOURCE 1

#define KAAPI_CACHE_LINE 64

/* in Byte */
#define KAAPI_KPROCESSOR_SIZE 4096

/* to be adapted on line */
#define KAAPI_ONELINE_DFG_THRESHOLD 1024

/* fwddecla */
struct kaapi_hashmap_t;

#include <errno.h>  // error code
#include <stdlib.h> // free, malloc

#include "config.h"
#include "kaapi_defs.h"
#include "kaapi_atomic.h"
#include "kaapi.h"
#include "kaapi_task.h"
#include "kaapi_memory.h"
#include "kaapi_perf.h"
#include "kaapi_workqueue.h"

#include "kaapi_wsprotocol.h"

#if defined(KAAPI_DEBUG)
#include <signal.h>
#include "kaapi_dbg.h"
#define KAAPI_ATOMIC_PRINT( inst ) \
{\
  kaapi_dbg_print_lock();\
  inst;\
  fflush(stdout); \
  kaapi_dbg_print_unlock();\
}
#else
#define KAAPI_ATOMIC_PRINT( inst ) 
#endif


/** Current implementation relies on isoaddress allocation to avoid
    communication during synchronization at the end of partitionning.
*/
//#define KAAPI_ADDRSPACE_ISOADDRESS 1


/** This is the new version on top of X-Kaapi
*/
extern const char* get_kaapi_version(void);
extern const char* get_kaapi_git_hash(void);
extern const char* get_kaapi_configure_log(void);


/* ================== Library initialization/terminaison ======================= */
/* Low level initialization based on value of kaapi_default_param to initialize
   submodules.
   kaapi_init is only a driver to get information from environment variables to
   setup kaapi_default_param fields.
*/
extern int kaapi_mt_init(int flag);
extern int kaapi_mt_hwdetect(void);
extern int kaapi_mt_registerproc(void);
extern int kaapi_mt_finalize(void);


/* Fwd declaration 
*/
struct kaapi_listrequest_t;

/* Fwd declaration
*/
struct kaapi_taskdescr_t;
struct kaapi_big_hashmap_t;
struct kaapi_version_t;
struct kaapi_metadata_info_t;
struct kaapi_mem_data_t;

/** Predefined value for KAAPI_DISPLAY_PERF
*/
typedef enum  {
  KAAPI_NO_DISPLAY_PERF     = 0,  /* do not display performance counters */
  KAAPI_DISPLAY_PERF_FULL   = 1,  /* display all counters, for each threads */
  KAAPI_DISPLAY_PERF_RESUME = 2,  /* display only cumulated counters per process */
  KAAPI_DISPLAY_PERF_GPLOT  = 3   /* display only gplot output */
} kaapi_display_perf_value_t;

/* \ingroup SCHED
 * Return a priority from ct path value
 */
typedef uint8_t (*kaapi_ct2prio_fnc_t)(uint64_t, struct kaapi_task_t*);


/* ========================================================================= */
/* Runtime environment, set through environement variable                    */
/* ========================================================================= */

typedef struct kaapi_request_t* (*kaapi_listrequest_iterator_get_fnc_t)(
  struct kaapi_listrequest_iterator_t* lrrange
);
typedef struct kaapi_request_t* (*kaapi_listrequest_iterator_next_fnc_t)(
  struct kaapi_listrequest_iterator_t* lrrange 
);
typedef int (*kaapi_listrequest_iterator_empty_fnc_t)(
  struct kaapi_listrequest_iterator_t* lrrange
);
/* may reclaims O(k) ops */
typedef int (*kaapi_listrequest_iterator_count_fnc_t)(
  struct kaapi_listrequest_iterator_t* lrrange
);

/* base structure that is embeded into specific list iterator / aggregation strategy */
typedef struct kaapi_listrequest_iterator_t {
  kaapi_listrequest_iterator_get_fnc_t   lri_get;
  kaapi_listrequest_iterator_next_fnc_t  lri_next;
  kaapi_listrequest_iterator_empty_fnc_t lri_empty;
  kaapi_listrequest_iterator_count_fnc_t lri_count;
} kaapi_listrequest_iterator_t;

#if defined(KAAPI_USE_PERFCOUNTER)
#include "kaapi_event.h"
#include "kaapi_trace.h"
#endif
#include "kaapi_machine.h"
#include "kaapi_sched.h"
#include "kaapi_hashmap.h"
#include "kaapi_format.h"

#if defined(KAAPI_USE_PERFCOUNTER)
#include "kaapi_event_recorder.h"
#endif


/** Definition of parameters for the runtime system
*/
typedef struct kaapi_rtparam_t {
  size_t                        stackblocsize;       /* default stack bloc size */
  size_t                        threadstacksize;     /* default pthread stack size */
  unsigned int                  syscpucount;         /* number of physical cpus of the system */
  unsigned int                  cpucount;            /* number of physical cpu used for execution */
  const char*                   cpuset_str;          /* string from getenv */
  const char*                   cpucount_str;        /* string from getenv */
  unsigned int                  gpucount;            /* number of physical gpu used for execution */
  kaapi_selectvictim_fnc_t      wsselect;            /* default method to select a victim */
  kaapi_localsteal_fnc_t        wslocalsteal;        /* method call when thread is idle before emitsteal */
  kaapi_sched_select_queue_fnc_t wspush;             /* call to select queue where to push task */
  kaapi_emitsteal_fnc_t	        emitsteal;
  kaapi_emitsteal_init_t        emitsteal_initctxt;  /* call to initialize the emitsteal ctxt */
  kaapi_emitsteal_dstor_t       emitsteal_dstorctxt; /* call to dstor ctxt */

#if defined(KAAPI_USE_PERFMODEL)
  unsigned int                  perfmodel;           /* enable performance model for prediction */
  unsigned int                  perfmodel_calibrate; /* enable time counters for the performance model */
  unsigned int                  noremotesteal;       /* no steal from remote-pushed tasks */
#endif

  unsigned int		              use_affinity;        /* use cpu affinity if CPUSET defined */
  kaapi_affinityset_t           affinity_set[KAAPI_HWS_LEVELID_MAX];
  kaapi_procbind_t              procbind;            /* */
  kaapi_places_part_t           places;              /* initial places */

  kaapi_ct2prio_fnc_t		        ctpriority;  /* use critical path priorities, if 0 no */
  kaapi_display_perf_value_t    display_perfcounter; /* see value above for kaapi_display_perf_value_t  */
#if defined(KAAPI_USE_CUDA)
   uint64_t		                  cudawindowsize;
   unsigned int		              cudapeertopeer;
#if defined(KAAPI_USE_CUPTI)
  uint64_t		                  cudastartuptime;
#endif
#endif
  uint64_t                      startuptime;         /* time at the end of kaapi_init */
  int                           alarmperiod;         /* period for alarm */
  uint64_t                      eventmask;           /* event mask */

  int                           papi_event_count;    /* !=0 if PAPI is configured and KAAPI_PERF_PAPIES is defined */
  int                           trace_recorder;      /* !=0 if trace recorder is registered */
} kaapi_rtparam_t;

extern kaapi_rtparam_t kaapi_default_param;




/* ===================== Helper task functions  ==================================== */
static inline const struct kaapi_format_t* _kaapi_task_getformat( const kaapi_task_t* task)
{
  const struct kaapi_format_t* fmt = 0;
  if ((task->u.s.flag & KAAPI_TASK_FLAG_DFGOK) == 0)
    fmt = kaapi_format_resolvebybody( task->body );
  else
    fmt = task->fmt;
  return fmt;
}


static inline size_t _kaapi_task_getsize( const kaapi_task_t* task, const struct kaapi_format_t* fmt)
{
  size_t size = 0;
  unsigned int count = kaapi_format_get_count_params(fmt, task->arg);
  unsigned int i;
  for (i=0; i<count; ++i)
  {
    if (KAAPI_ACCESS_GET_MODE(kaapi_format_get_mode_param(fmt, i, task->arg)) == KAAPI_ACCESS_MODE_V)
      continue;
    kaapi_memory_view_t view;
    kaapi_format_get_view_param(fmt, i, task->arg, &view);
    size += kaapi_memory_view_size(&view);
  }
  return size;
}

/* ========================================================================= */
/* */
extern uint64_t kaapi_perf_thread_delayinstate(struct kaapi_processor_t* kproc);


/* ========== MACHINE DEPEND DATA STRUCTURE =========== */


/** The Kaapi global team: master team creates by the main thread
*/
extern kaapi_team_t* kaapi_global_team;

/** Destroy a request
    A posted request could not be destroyed until a reply has been made
*/
#define kaapi_request_destroy( kpsr ) 

static inline kaapi_processor_id_t kaapi_request_getthiefid(kaapi_request_t* r)
{ return (kaapi_processor_id_t) r->ident; }

/** Return the request status
  \param pksr kaapi_request_t
  \retval KAAPI_REQUEST_S_OK sucessfull steal operation
  \retval KAAPI_REQUEST_S_NOK steal request has failed
  \retval KAAPI_REQUEST_S_ERROR error during steal request processing
*/
static inline int kaapi_request_status_get( kaapi_atomic_t* status ) 
{ return KAAPI_ATOMIC_READ(status); }

/** Return the request status
  \param pksr kaapi_request_t
  \retval KAAPI_REQUEST_S_OK sucessfull steal operation
  \retval KAAPI_REQUEST_S_NOK steal request has failed
  \retval KAAPI_REQUEST_S_ERROR error during steal request processing
*/
static inline int kaapi_request_status( kaapi_request_t* kr ) 
{ return kaapi_request_status_get(&kr->status); }

/** Return true iff the request has a reply
  \param kr kaapi_request_t
  \retval KAAPI_REQUEST_S_OK sucessfull steal operation
  \retval KAAPI_REQUEST_S_NOK steal request has failed
  \retval KAAPI_REQUEST_S_ERROR steal request has failed to be posted because the victim refused request
*/
static inline int kaapi_request_status_test( kaapi_atomic_t* status )
{ return KAAPI_ATOMIC_READ(status) != KAAPI_REQUEST_S_POSTED; }

/** Return true iff the request has been posted
  \param kr kaapi_request_t
  \retval KAAPI_REQUEST_S_OK sucessfull steal operation
  \retval KAAPI_REQUEST_S_NOK steal request has failed
  \retval KAAPI_REQUEST_S_ERROR steal request has failed to be posted because the victim refused request
*/
static inline int kaapi_request_test( kaapi_request_t* kr )
{ return kaapi_request_status_test(&kr->status); }

/** Return true iff the request is a success steal
  \param kr kaapi_request_t
*/
static inline int kaapi_request_ok( kaapi_request_t* kr )
{ return kaapi_request_status(kr) != KAAPI_REQUEST_S_NOK; }

/** Return the data associated with the reply
  \param kr kaapi_request_t
*/
static inline void kaapi_request_syncdata( kaapi_request_t* kr ) 
{ 
  kaapi_mem_barrier();
}

/* ======================== Perf counter interface: machine dependent ========================*/
#include "kaapi_trace.h"
#include "kaapi_event_recorder.h"


#if defined(KAAPI_DEBUG)
#if defined(KAAPI_USE_SCHED_AFFINITY)
/* output bits string */
extern char* kaapi_cpuset2string( int nproc, const cpu_set_t* cupset );
/* output bits string, reentrant version */
extern char* kaapi_cpuset2string_r( char* buffer, int size, int nproc, const cpu_set_t* cupset );
#endif

/* ======================== MACHINE DEPENDENT FUNCTION THAT SHOULD BE DEFINED ========================*/
/* ........................................ PUBLIC INTERFACE ........................................*/
extern void* kaapi_alloc_protect( size_t size );
extern void kaapi_free_protect( void* p );
extern void* kaapi_realloc_protect(void *ptr, size_t size);

/* Signal handler to dump the state of the internal kprocessors
   This signal handler is attached to SIGALARM when KAAPI_DUMP_PERIOD env. var. is defined.
*/
extern void _kaapi_signal_dump_state(int);

/* Signal handler to print the backtrace
   This signal handler is attached to:
    - SIGABRT
    - SIGTERM
    - SIGSEGV
    - SIGFPE 
    - SIGILL
    If the library is configured with --with-perfcounter, then the function call _kaapi_signal_dump_counters.
*/
extern void _kaapi_signal_dump_backtrace(int, siginfo_t *si, void *unused);
#endif


#if defined(KAAPI_USE_PERFCOUNTER)
/* Signal handler attached to:
    - SIGINT
    - SIGQUIT
    - SIGABRT
    - SIGTERM
    - SIGSTOP
  when the library is configured with --with-perfcounter in order to flush some counters.
*/
extern void _kaapi_signal_dump_counters(int);
#endif


#if defined(__cplusplus)
}
#endif

#endif
