/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com / fabien.lementec@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef _KAAPI_PERF_H
#define _KAAPI_PERF_H 1


#if defined(__cplusplus)
extern "C" {
#endif
/* ========================================================================= */
/* Perf counter                                                              */
/* ========================================================================= */
/** \ingroup PERF
    Performace counters. Please read documentation (for Marie :-)).
    The runtime instrumentation allows to compute performance counters at runtime.
    The set of counters is decomposed in two classes:
    - software counters 
    - hardware counters using PAPI support to read and accumulate them.
    Each counters has two values maintains by the instrumentation module depending
    if the incrementation of the counters occurs when the core performs application code (user)
    or when runtime functions are executed (sys).
    
    Some general remarks: all times are in ns, excepted if specified otherwise
*/
#define KAAPI_PERF_ID_USER_POS (31)
#define KAAPI_PERF_ID_USER_MASK (1 << KAAPI_PERF_ID_USER_POS)

#define KAAPI_PERF_ID(U, I)  (KAAPI_PERF_ID_ ## I | (U) << KAAPI_PERF_ID_USER_POS)
#define KAAPI_PERF_ID_USER(I) KAAPI_PERF_ID(1, I)
#define KAAPI_PERF_ID_PRIV(I) KAAPI_PERF_ID(0, I)


/* count number of executed tasks */
#define KAAPI_PERF_ID_TASKS         0  

/* count number of successful steal requests */
#define KAAPI_PERF_ID_STEALREQOK    1  

/* count number of steal requests emitted */
#define KAAPI_PERF_ID_STEALREQ      2  

/* count number of steal operation to reply to requests */
#define KAAPI_PERF_ID_STEALOP       3  

/* count number of receive steal requests */
#define KAAPI_PERF_ID_STEALIN       4  

/* count number of suspended operation */
#define KAAPI_PERF_ID_SUSPEND       5  

/* computation time for application (usr), or for scheduling (sys) */
#define KAAPI_PERF_ID_T1            6  

/* nano second passed in synchronization (usr=application, sys=runtime) 
   - only time spent in kaapi_sched_lock, kaapi_sched_waitlock and
   for time acquision during work stealing are reported.
   - use of kaapi_atomic_lock does not report time.
*/
#define KAAPI_PERF_ID_TSYNC         7  

/* nano second of preempt time */
#define KAAPI_PERF_ID_TPREEMPT      8  

/* time to compute task lists in ns */
#define KAAPI_PERF_ID_TASKLISTCALC  9  

#define KAAPI_PERF_ID_COMM_H2D	    10 /* host to device transfers */
#define KAAPI_PERF_ID_COMM_D2H      11 /* device to host transfers */
#define KAAPI_PERF_ID_COMM_D2D      12 /* device to device transfers */
#define KAAPI_PERF_ID_CACHE_HIT	    13 /* GPU cache hit */
#define KAAPI_PERF_ID_CACHE_MISS    14 /* GPU cache miss */
#define KAAPI_PERF_ID_STACKSIZEBLOC 15 /* bytes of memory bloc stack allocated */
#define KAAPI_PERF_ID_CONFLICTPOP   16 /* conflictual frame pop */

/* mark end of software counters */
#define KAAPI_PERF_ID_ENDSOFTWARE   17

#define KAAPI_PERF_ID_PAPI_BASE    (KAAPI_PERF_ID_ENDSOFTWARE)
#define KAAPI_PERF_ID_PAPI_0       (KAAPI_PERF_ID_PAPI_BASE + 0)
#define KAAPI_PERF_ID_PAPI_1       (KAAPI_PERF_ID_PAPI_BASE + 1)
#define KAAPI_PERF_ID_PAPI_2       (KAAPI_PERF_ID_PAPI_BASE + 2)

#define KAAPI_PERF_ID_PAPI_MAX     (KAAPI_PERF_ID_PAPI_2 - KAAPI_PERF_ID_PAPI_BASE + 1)
#define KAAPI_PERF_ID_MAX          (KAAPI_PERF_ID_PAPI_2 + 1)
#define KAAPI_PERF_ID_ALL           KAAPI_PERF_ID_MAX

#define KAAPI_PERF_USR_COUNTER      0x1  /* for kaapi_perf_accum_counters or kaapi_perf_read_counters*/
#define KAAPI_PERF_SYS_COUNTER      0x2  /* could be ored */

/* counter type
*/
typedef int64_t kaapi_perf_counter_t;

/* value
*/
typedef unsigned int kaapi_perf_id_t;
typedef struct kaapi_perf_idset_t
{
  unsigned int count;
  unsigned char idmap[KAAPI_PERF_ID_MAX];
} kaapi_perf_idset_t;

#define KAAPI_PERF_IDSET_SINGLETON(I) {1, {I}}
#define KAAPI_PERF_IDSET_TASKS ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_TASKS)
#define KAAPI_PERF_IDSET_STEALREQOK ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_STEALREQOK)
#define KAAPI_PERF_IDSET_STEALREQ ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_STEALREQ)
#define KAAPI_PERF_IDSET_STEALOP ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_STEALOP)
#define KAAPI_PERF_IDSET_SUSPEND ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_SUSPEND)
#define KAAPI_PERF_IDSET_TIDLE ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_TIDLE)
#define KAAPI_PERF_IDSET_TPREEMPT ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_TPREEMPT)
#define KAAPI_PERF_IDSET_PAPI_0 ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_PAPI_0)
#define KAAPI_PERF_IDSET_PAPI_1 ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_PAPI_1)
#define KAAPI_PERF_IDSET_PAPI_2 ((kaapi_perf_idset_t*)(uintptr_t)KAAPI_PERF_ID_PAPI_2)


/* for perf_regs access: SHOULD BE 0 and 1 
   All counters have both USER and SYS definition (sys == program that execute the scheduler).
   * KAAPI_PERF_ID_T1 is considered as the T1 (computation time) in the user space
   and as TSCHED, the scheduling time if SYS space. In workstealing litterature it 
   is also named Tidle.
*/
#define KAAPI_PERF_USER_STATE       0
#define KAAPI_PERF_SCHEDULE_STATE   1


/* idset */
extern void kaapi_perf_idset_zero(kaapi_perf_idset_t*);
extern void kaapi_perf_idset_add(kaapi_perf_idset_t*, kaapi_perf_id_t);

/* global access */
extern void _kaapi_perf_accum_counters(const kaapi_perf_idset_t* idset, int isuser, kaapi_perf_counter_t* counter);
extern void _kaapi_perf_read_counters(const kaapi_perf_idset_t* idset, int isuser, kaapi_perf_counter_t* counter);

/* global access, user functions */
static inline void kaapi_perf_accum_counters(const kaapi_perf_idset_t* idset, kaapi_perf_counter_t* counter)
{ _kaapi_perf_accum_counters(idset, KAAPI_PERF_USR_COUNTER, counter); }

/* */
static inline void kaapi_perf_read_counters(const kaapi_perf_idset_t* idset, kaapi_perf_counter_t* counter)
{ _kaapi_perf_read_counters(idset, KAAPI_PERF_USR_COUNTER, counter); }

/* local access, internal functions */
extern void _kaapi_perf_read_register(const kaapi_perf_idset_t* idset, int isuser, kaapi_perf_counter_t* counter);
extern void _kaapi_perf_accum_register(const kaapi_perf_idset_t* idset, int isuser, kaapi_perf_counter_t* accum);

/* local access, user functions */
static inline void kaapi_perf_read_register(const kaapi_perf_idset_t* idset, kaapi_perf_counter_t* counter)
{ _kaapi_perf_read_register(idset, KAAPI_PERF_USR_COUNTER, counter); }

static inline void kaapi_perf_accum_register(const kaapi_perf_idset_t* idset, kaapi_perf_counter_t* accum)
{ _kaapi_perf_accum_register(idset, KAAPI_PERF_USR_COUNTER, accum); }

/* utility */
extern const char* kaapi_perf_id_to_name(kaapi_perf_id_t);
extern size_t kaapi_perf_counter_num(void);

extern void _kaapi_perf_read_counters(const kaapi_perf_idset_t* idset, int isuser, kaapi_perf_counter_t* counter);

/* force checkpoint of possible counters of the current running thread and put it in trace file if used */
extern void kaapi_perf_thread_checkpoint(void);

/* ========================================================================= */
/* Perf counter region: experimental feature                                 */
/* ========================================================================= */

/* Start a new region. Only 4 char of the name will be recorderd in the trace
   The current implementation assume that the parallel program is a sequence of
   regions. A region cannot be embded into an other region.
   At the start of a region, all threads record their performance counters to the
   trace file.
   At the stop of a region, all threads record again their performance counters.
*/
extern void kaapi_perf_start_region( const char* name );

/*
*/
extern void kaapi_perf_stop_region( );

#if defined(__cplusplus)
}
#endif

#endif
