/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com / fabien.lementec@imag.fr
** vincent.danjean@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef _KAAPI_MEMORY_H
#define _KAAPI_MEMORY_H 1


#if defined(__cplusplus)
extern "C" {
#endif


/* ========================== NUma associated functions ==================== */
typedef uint32_t kaapi_numaid_t;
typedef uint64_t kaapi_numaset_t;

/* Initialize
*/
extern int kaapi_numa_initialize(void);

/*
*/
extern int kaapi_numa_finalize(void);

/*
*/
extern void* kaapi_numa_alloc(size_t size);

/*
*/
extern int kaapi_numa_free(void* ptr, size_t size);

/*
*/
void* kaapi_numa_allocate_bloc1dcyclic(kaapi_numaset_t ldset, size_t size, size_t blocsize);

/*
*/
extern void* kaapi_numa_allocate_bloc2dcyclic(
  kaapi_numaset_t numaset, size_t nsetx, size_t nsety, /* grid of numa node */
  size_t n, size_t m, size_t wordsize,    /* matrix */
  size_t nbx,                                          /* number of blocs / x */
  size_t nby                                           /* number of blocs / y */
);

/*
*/
extern int kaapi_numa_bind_bloc2dcyclic(
  kaapi_numaset_t numaset, size_t nsetx, size_t nsety, /* grid of numa node */
  const void* addr, size_t n, size_t m, size_t ld, size_t wordsize,    /* matrix */
  size_t nbx,                                          /* number of blocs / x */
  size_t nby                                           /* number of blocs / y */
);

/*
*/
extern int kaapi_numa_bind_bloc1dcyclic(
  kaapi_numaset_t numaset, const void* addr, size_t size, size_t blocsize
);

/*
*/
extern int kaapi_numa_bind(kaapi_numaid_t node, const void* addr, size_t size);

/*
*/
extern kaapi_numaid_t kaapi_numa_getpage_id(const void* addr);



/* ============================ Memory type ================================= */
/** Copy a view of a data to an other view in a remote address space.
    Source and destination memory region cannot overlap.
    \retval 0 in case of success
*/
extern int kaapi_memory_copy( 
  kaapi_pointer_t dest, const kaapi_memory_view_t* view_dest,
  kaapi_pointer_t src,  const kaapi_memory_view_t* view_src 
);


/** Copy a view of a data to an other view in a remote address space.
    Source and destination memory region cannot overlap.
    \retval 0 in case of success
*/
extern int kaapi_memory_asyncopy( 
  kaapi_pointer_t dest, const kaapi_memory_view_t* view_dest,
  const kaapi_pointer_t src, const kaapi_memory_view_t* view_src,
  void (*callback)(void*), void* argcallback
);


/** assume that now the view points to a new allocate view
*/
static inline void kaapi_memory_view_reallocated( kaapi_memory_view_t* kmv )
{
  switch (kmv->type) 
  {
    case KAAPI_MEMORY_VIEW_1D: return;
    case KAAPI_MEMORY_VIEW_2D: kmv->lda = kmv->size[1]; return;
    default:
      kaapi_assert(0);
      break;
  }
}


/* Representation of address space id:
   - 256 numa node id
   - 256 local identifier
   - (0x00FFFFFFFFFF0000ULL>>16) global identifier
   - 16 bit for coding architecture
   numa node id is considered as sub domain of the address space id.
   For instance on one multicore machine there is one address space id
   with several numa node (up to 256).
*/
#define KAAPI_ASID_MASK_NUID  0x00000000000000FFULL /* shift = 0 */
#define KAAPI_ASID_MASK_LID   0x000000000000FF00ULL /* shift = 8 */
#define KAAPI_ASID_MASK_GID   0x00FFFFFFFFFF0000ULL /* shift = 16 */
#define KAAPI_ASID_MASK_ARCH  0xF000000000000000ULL /* shift = 56 */

/** Return the gid of the address space
    \param kasid [IN] an address space identifier
    \return the gid encoded into the address space identifier
*/
static inline kaapi_globalid_t kaapi_memory_address_space_getgid( kaapi_address_space_id_t kasid )
{ return (kaapi_globalid_t)((kasid & KAAPI_ASID_MASK_GID)>> 16ULL); }

/** Return the type of the address space location.
    \param kasid [IN] an address space identifier
    \return the type encoded into the address space identifier
*/
static inline int kaapi_memory_address_space_gettype( kaapi_address_space_id_t kasid )
{ return (int)((kasid &KAAPI_ASID_MASK_ARCH) >> 56UL); }

/* return the address space identifier
*/
static inline kaapi_address_space_id_t kaapi_pointer2asid(kaapi_pointer_t p)
{ return p.asid; }

/* return the location (gid) where the pointer points.
*/
static inline kaapi_globalid_t kaapi_pointer2gid(kaapi_pointer_t p)
{ return kaapi_memory_address_space_getgid(p.asid); }

#if defined(__cplusplus)
}
#endif

#endif
