/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"
#include <errno.h>

#if defined(KAAPI_DEBUG)
#include <unistd.h>
#include <sys/mman.h>
#endif

#define KAAPI_ALIGNDOWN(addr,a) (((uintptr_t)(addr)) & ~((a)-1))

kaapi_stack_bloc_t* kaapi_processor_alloc_stackbloc(kaapi_processor_t* kproc)
{

  kaapi_stack_bloc_t* retbloc;
  size_t sizealloc = kaapi_default_param.stackblocsize;
  kaapi_assert_debug( sizealloc >=  KAAPI_STACKBLOCSIZE);
  kaapi_assert_debug( sizealloc % getpagesize() == 0);
  kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
  if (kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc !=0)
  {
    retbloc = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = retbloc->next;
    kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

    retbloc->next = 0;
    return retbloc;
  }
  kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

  int err = posix_memalign((void**)&retbloc, KAAPI_STACKBLOCSIZE, sizealloc );
  kaapi_assert_debug_m(err ==0, "stack alloc");
  if (err !=0)
    return 0;
  retbloc->size = kaapi_default_param.stackblocsize - offsetof(kaapi_stack_bloc_t,data);
  retbloc->next = 0;
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    KAAPI_PERF_REG(kproc->perf, KAAPI_PERF_ID_STACKSIZEBLOC) += 1;
#endif
  return retbloc;
}


/* no lock */
static inline void _kaapi_processor_dealloc_stackbloc(
    kaapi_processor_t* kproc, 
    kaapi_stack_bloc_t* bloc
)
{
  bloc->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
  kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = bloc;
}

/* */
void kaapi_processor_dealloc_stackbloc(
    kaapi_processor_t* kproc, 
    kaapi_stack_bloc_t* bloc
)
{
  kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
  _kaapi_processor_dealloc_stackbloc(kproc, bloc);
  kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
}


/*
*/
int kaapi_stack_init( kaapi_processor_t* kproc, kaapi_stack_t* stack )
{
  char* first_data;
  kaapi_frame_t* frame;

  stack->bloc_task = kaapi_processor_alloc_stackbloc(kproc);
  if (stack->bloc_task ==0)
    KAAPI_RETURN_ERROR(ENOMEM,0);
  stack->bloc_task0 = stack->bloc_task;
  stack->bloc_data = kaapi_processor_alloc_stackbloc(kproc);
  if (stack->bloc_data ==0)
    KAAPI_RETURN_ERROR(ENOMEM,0);
  stack->bloc_data0 = stack->bloc_data;

  first_data = kaapi_firstin_stack_bloc(stack->bloc_data, char);

  /* this frame should never be deleted */
  frame = (kaapi_frame_t*)first_data;
  first_data += sizeof(kaapi_frame_t);
  stack->top_frame = frame;
  _kaapi_frame_init( frame );

  stack->thread.sp_data      = first_data;
  stack->thread.sp           = kaapi_firstin_stack_bloc(stack->bloc_task, kaapi_task_t);
  stack->pc                  = stack->thread.sp;
  stack->unlink              = frame;

  stack->popfp               = 1;
  stack->thieffp             = 0;
  stack->cnt_task            = 0;

  return kaapi_atomic_initlock( &stack->lock);
}


/*
*/
int kaapi_stack_reset(kaapi_stack_t* stack )
{
  char* first_data;
  kaapi_frame_t* frame;
  kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;

  /* */
  kaapi_assert_debug( stack->top_frame->next == 0 );

  /* pop all frames until stack->top_frame */
  kaapi_assert_debug( stack->bloc_data->next == 0  );
  kaapi_assert_debug( stack->bloc_task->next == 0  );

  /* the first pushed remains in the stack */
  kaapi_assert_debug( stack->pc == stack->thread.sp );
  kaapi_assert_debug( stack->popfp == 1 );

  /* free bloc between bloc_task0 and bloc_task */
  if ((stack->bloc_task != stack->bloc_task0) || (stack->bloc_data != stack->bloc_data0))
  {
    kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
    if (stack->bloc_task != stack->bloc_task0)
    {
      stack->bloc_task->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
      kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = stack->bloc_task0->next;
      stack->bloc_task0->next = 0;
    }
    if (stack->bloc_data != stack->bloc_data0)
    {
      stack->bloc_data->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
      kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = stack->bloc_data0->next;
      stack->bloc_data0->next = 0;
    }
    kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
    stack->bloc_task = stack->bloc_task0;
    stack->bloc_data = stack->bloc_data0;
  }

  first_data = kaapi_firstin_stack_bloc(stack->bloc_data, char);
  frame = (kaapi_frame_t*)first_data;
  first_data += sizeof(kaapi_frame_t);
  kaapi_assert_debug( stack->top_frame == frame );
  _kaapi_frame_init(frame);
/*
  if (frame->rdlist && (frame->flag & KAAPI_FRAME_FLAG_DFG_OK))
    kaapi_hashmap_clear( &frame->rdlist->ctxt.access_to_gd );
*/

  /* ordered like this to allow concurrent update */
  stack->pc                = 0;
  kaapi_writemem_barrier();

  stack->thread.sp_data    = first_data;
  stack->thread.sp         = kaapi_firstin_stack_bloc(stack->bloc_task, kaapi_task_t);
  stack->pc                = stack->thread.sp;
  stack->unlink            = frame;

  kaapi_writemem_barrier();

  stack->pc                = stack->thread.sp;
  stack->cnt_task          = 0;

  return 0;
}


/*
*/
int kaapi_stack_destroy(kaapi_processor_t* kproc, kaapi_stack_t* stack )
{
  if (stack->thread.sp != stack->pc)
    return EBUSY;

  /* free bloc between bloc_task0 and bloc_task */
  if ((0 != stack->bloc_task0) || (0 != stack->bloc_data0))
  {
    kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

    stack->bloc_task->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = stack->bloc_task0;

    stack->bloc_data->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = stack->bloc_data0;

    kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

    stack->bloc_task = stack->bloc_task0 = 0;
    stack->bloc_data = stack->bloc_data0 = 0;
  }

  stack->top_frame = 0;
  
  return kaapi_atomic_destroylock( &stack->lock);
}


/* Not complete version. Return >= 32 if new bloc was allocated.
   Does not correspond to correct frame definition of tasks. See sched_sync.
*/
uint64_t kaapi_frame_count_task( const kaapi_stack_t* stack, const kaapi_frame_t* frame )
{
  if (stack->bloc_task != stack->bloc_task0)
    return 32;
  return stack->thread.sp - stack->pc; 
 // return frame->cnt_task + (_kaapi_end_blocaddr(stack->pc,kaapi_task_t*)-1 - stack->pc);
}


/*
*/
int kaapi_stack_iterator_init(
  kaapi_stack_iterator_t* iter,
  kaapi_stack_t* stack,
  kaapi_frame_t* frame
)
{
  if (frame->next ==0)
  { /* means current runnng frame */
    iter->last  = stack->thread.sp;
    iter->first = stack->pc;
  }
  else
  { /* means pushed frame */
    iter->last  = frame->save_thread.sp;
    iter->first = frame->save_pc;
  }
  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(iter->first, kaapi_stack_bloc_t*);
  kaapi_task_t* sp_bloc = bloc->next == 0 ?  iter->last : _kaapi_end_blocaddr(iter->first,kaapi_task_t*)-1;
  if (iter->first == sp_bloc)
  {
    kaapi_stack_bloc_t* nextbloc = bloc->next;
    if (nextbloc ==0) 
    {
      iter->first = iter->last;
      return 0;
    }
    iter->first = kaapi_firstin_stack_bloc(nextbloc, kaapi_task_t);
    bloc = _kaapi_start_blocaddr( iter->last, kaapi_stack_bloc_t* );
    /* case where new bloc is inserted and linked, but sp not yet incremented */
    if ((bloc == nextbloc) && (iter->first > iter->last))
      iter->first = iter->last;
    /* assume iter->first != sp_bloc */
    kaapi_assert_debug( 
       iter->first != (bloc->next == 0 ?  iter->last : _kaapi_end_blocaddr(iter->first,kaapi_task_t*)-1)
    );
  }
  return 0;
}


/* Pass to the next task. 
*/
void kaapi_stack_iterator_next( kaapi_stack_iterator_t* iter )
{
  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(iter->first, kaapi_stack_bloc_t*);
  kaapi_task_t* sp_bloc = bloc->next == 0 ?  iter->last : _kaapi_end_blocaddr(iter->first,kaapi_task_t*)-1;
  ++iter->first;
  if (iter->first == iter->last) 
    return;
  if (iter->first == sp_bloc)
  {
    kaapi_stack_bloc_t* nextbloc = bloc->next;
    if (nextbloc ==0) return;
    bloc = _kaapi_start_blocaddr( iter->last, kaapi_stack_bloc_t* );
    iter->first = kaapi_firstin_stack_bloc(nextbloc, kaapi_task_t);
    /* case where new bloc is inserted and linked, but sp not yet incremented */
    if ((bloc == nextbloc) && (iter->first > iter->last))
      iter->first = iter->last;
  }
}


/* iterates over iter->first downto iter->last
*/
int kaapi_stack_reverse_iterator_init(
  kaapi_stack_iterator_t* iter,
  kaapi_stack_t*          stack,
  kaapi_frame_t*          frame
)
{
  if (frame->next ==0)
  { /* means current runnng frame */
    iter->first = stack->thread.sp;
    iter->last  = stack->pc-1;
  }
  else
  { /* means pushed frame */
    iter->first = frame->save_thread.sp;
    iter->last  = frame->save_pc-1;
  }

  /* correct first to point to the first task in the iteration == last pushed task == sp-1 */
  kaapi_stack_bloc_t* sp_bloc = _kaapi_start_blocaddr(iter->first, kaapi_stack_bloc_t*);
  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(iter->first-1, kaapi_stack_bloc_t*);
  if (bloc != sp_bloc)
    iter->first = (kaapi_task_t*)sp_bloc->save_ptr;
  else
    --iter->first; 

  return 0;
}


/* Pass to the prev task.
*/
void kaapi_stack_reverse_iterator_next( kaapi_stack_iterator_t* iter )
{
  kaapi_stack_bloc_t* sp_bloc = _kaapi_start_blocaddr(iter->first, kaapi_stack_bloc_t*);
  if (--iter->first == iter->last)
    return;

  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(iter->first,kaapi_stack_bloc_t*);
  if (sp_bloc != bloc)
  if (iter->first == iter->last)
    iter->first = (kaapi_task_t*)sp_bloc->save_ptr;
}
