/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com / fabien.lementec@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#ifndef _KAAPI_MT_MACHINE_H_
#define _KAAPI_MT_MACHINE_H_ 1

#include "config.h"
#include <pthread.h>
#include <stddef.h> // offset of

#ifdef __linux__
#  ifdef HAVE_UCONTEXT_H
#    define KAAPI_USE_UCONTEXT
#  elif HAVE_SETJMP_H
#    error "Not implemented yet"
#    define KAAPI_USE_SETJMP
#  endif
#endif

#if defined(KAAPI_USE_SCHED_AFFINITY)
#include <sched.h>
#endif

#ifdef __APPLE__
#  ifdef HAVE_SETJMP_H
#    define KAAPI_USE_SETJMP
#  elif HAVE_UCONTEXT_H
#    define KAAPI_USE_UCONTEXT
#  endif
#endif


#if defined(KAAPI_USE_SETJMP)
#  include <setjmp.h>
#elif defined(KAAPI_USE_UCONTEXT)
#  include <ucontext.h>
#endif

#if defined(KAAPI_USE_NUMA)
#  include <numa.h>
#  include <numaif.h>
#  define KAAPI_KPROCESSOR_ALIGNMENT_SIZE 4096  /* page size */
#else
#  define KAAPI_KPROCESSOR_ALIGNMENT_SIZE KAAPI_CACHE_LINE
#endif

#include "kaapi_hashmap.h"
#include "kaapi_memory.h"

struct kaapi_team_t;
struct kaapi_context_t;
struct kaapi_request_node_t;
struct kaapi_place_t;
struct kaapi_foreach_globalwork_t;


/* ========================================================================== */
/* ========================================================================== */
#if defined(KAAPI_USE_SCHED_AFFINITY)
typedef cpu_set_t   kaapi_place_t;
#else
typedef void*       kaapi_place_t;
#endif
typedef struct {
  kaapi_place_t* first;
  unsigned long  len;
} kaapi_places_part_t;

/* ========================================================================== */
/*  ProcInfo struct is used to initialize a set of thread with 
    ressource constrainst such as cpuid.
    This structure is used to start thread after field team was set.
    run_processor is the subroutine start with thre thread. If 0 then Kaapi
    start a default subroutine.
*/ 
/* ========================================================================== */
/* */
typedef struct kaapi_procinfo_t
{
  struct kaapi_procinfo_t* next;
  kaapi_places_part_t  place_part;
  long                 place_off;   /* in place_part */
  unsigned int         proc_type;
  unsigned int         proc_index;
  unsigned int         reuse;
  void*                arg_runprocessor[2];
  struct kaapi_team_t* team;
} kaapi_procinfo_t;


/* ========================================================================== */
/* reserved field in kproc structure, only used if lib configured with perfcounter
*/
struct kaapi_event_buffer_t;

/*
*/
typedef kaapi_perf_counter_t kaapi_threadperf_counter_t[2][KAAPI_PERF_ID_MAX];

/** lfree data structure
*/
typedef struct kaapi_lfree
{
  kaapi_list_t            list;
  int                     _sizelfree;
} kaapi_lfree_t;

#define KAAPI_CACHE_LINE_SIZE 64


/* ========================================================================= */
/* Information about memory hierarchy                                               */
/* ========================================================================= */
/** nodeinfo is the static object attached to a localitydomain at given level
*/
typedef struct kaapi_nodeinfo_t {
  kaapi_ldid_t               lid;
#if defined(KAAPI_USE_SCHED_AFFINITY)
  cpu_set_t                  cpuset;
#endif
  kaapi_lock_t               lock __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));
  kaapi_lfree_t              lfree;                     /* queue of free context shared at nodeinfo level */
  kaapi_stack_bloc_t*        freebloc;                  /* stack bloc allocator, shared at nodeinfo level */
} kaapi_nodeinfo_t __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));

typedef struct kaapi_affinityset_t {
  kaapi_hws_levelid_t        hid;
  int                        count;
  kaapi_nodeinfo_t*          nis;
} kaapi_affinityset_t;

/** Dynamic view of nodeinfo with kids as reference.
    Each kprocessor have access to a hierarchy of localitydomain.
    See _kaapi_init_ld for initialization of fields
*/
typedef struct kaapi_localitydomain_t {
  kaapi_ldid_t               lid;
  int                        capacity;       /* size of kids */
  int                        count;          /* number of workers in the domain */
  struct kaapi_processor_t** kids;           /* array of size capacity workers */
  kaapi_lock_t               lock_r          /* used for locking queue. lock consummer if spsc */
    __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));
  kaapi_lock_t               lock_up;        /* for HWS strategy: lock by the first idle cores */
  kaapi_queue_t              queue;          /* queue of tasks */
  kaapi_queue_t              private_queue;  /* private queue of tasks to kproc of the ld */
} kaapi_localitydomain_t __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));

/*
*/
typedef struct kaapi_localitydomainset_t {
  uint64_t                 bitmap; /* bit i ==1 -> lds[i] not nul for NUMA level only */
  int                      count;  /* size of lds */
  kaapi_localitydomain_t** lds;    /* array of domains */
} kaapi_localitydomainset_t;


/* ========================================================================= */
/* Barrier                                                                   */
/* ========================================================================= */
#define KAAPI_BAR_CYCLES 3
typedef struct kaapi_barrier_t {
  kaapi_atomic_t cycle __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));
  kaapi_atomic_t wait_cycle __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));
  int term __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE)));
  char __attribute__ ((aligned (KAAPI_CACHE_LINE_SIZE))) count[KAAPI_BAR_CYCLES * KAAPI_CACHE_LINE_SIZE];
} kaapi_barrier_t;


/* ========================================================================= */
/* Team a set of kprocessors                                                 */
/* ========================================================================= */

typedef struct kaapi_team_t {
  kaapi_thread_id_t          tid;                       /* id of the creator thread */
  int                        count;                     /* threads */
  int                        teamid;
  int volatile               isterm;
  int                        flag;                      /* see kaapi_teamflag_t */
  
  kaapi_barrier_t            barrier __attribute__((aligned (KAAPI_CACHE_LINE)));
  kaapi_lock_t               lock;

  struct kaapi_processor_t** all_kprocessors;

  /* OMP work share region */
  struct workshare_t {
    kaapi_lock_t             lock;                      /* for omp lock similar construct */
    kaapi_atomic_t           single;                    /* for omp single similar construct */
    void*                    single_data;               /* for libGOMP copyprivate */
    kaapi_atomic_t           forcst;                    /* for omp for similar construct */
    struct kaapi_foreach_globalwork_t* gwork;           /* for omp for construct, terminaison */
    kaapi_atomic64_t         ordered_index;             /* ordered clause management of loop */
    double                   data[8];                   /* used by omp to avoid small allocation */
  } ws __attribute__((aligned (KAAPI_CACHE_LINE)));

  kaapi_selectvictim_fnc_t   fnc_select;                /* function to select a victim */
  kaapi_localsteal_fnc_t     fnc_localsteal;            /* function called for local steal first */
  kaapi_emitsteal_fnc_t	     emitsteal;                 /* he WS algorithm */

  struct kaapi_team_t*       parent_team;
  kaapi_localitydomainset_t  localitydomains[KAAPI_HWS_LEVELID_MAX];


  struct kaapi_processor_t*  save_master;               /* current kproc before team creation */
  struct kaapi_thread_register_t*reg0;                  /* state of the thread register before team init */

  kaapi_atomic_t             started;
  int                        thread_started;            /* cache for the main team */
  kaapi_procinfo_t*          all_descriptors;           /* to start thread */

#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_threadperf_counter_t*	perf_regs;                /* of size count*2 for 2 modes per thread*/
#endif
} kaapi_team_t __attribute__((aligned (KAAPI_CACHE_LINE)));


enum kaapi_teamflag_t {
  KAAPI_TEAM_NOCACHE
};

/* ========================================================================= */
/* Context data structure                                                    */ 
/* ========================================================================= */
/** The context data structure
    This data structure should be extend in case where the C-stack  is required to be suspended 
    and resumed.

    WARNING: stack should be the first field of the data structure in order to be able 
    to recover in the public kaapi_stack from public kaapi_thread_t*
*/
typedef struct kaapi_context_t {
  kaapi_stack_t                    stack;
  struct kaapi_processor_t*        proc;
  int                              unstealable;       /* !=0 -> cannot be stolen */

  /* OMP support */
  struct thread_specific_ws {                         /* thread specific work share info */
    int                            count_single;      /* for single as omp single construct */
    int                            count_for;         /* for single as omp single construct */
    int                            count_for_end;     /* for single as omp single construct */
    struct {
      struct kaapi_foreach_work_t* volatile rt_ctxt;  /* for GOMP support */
      long long                    first;
      long long                    last;
      long long                    incr;
      long long                    curr_start;        /* for ordered OMP support */
      long long                    curr_end;
    } for_data;
  } ws;
  void*                            komp_icvs;
  long                             place_off;
  kaapi_places_part_t              place_part;

  kaapi_task_t                     stask;
  kaapi_address_space_id_t         asid;           /* the address where is the thread */
} __attribute__((aligned (KAAPI_CACHE_LINE))) kaapi_context_t;

/* helper function */
#define kaapi_context2thread(ctxt)       (&(ctxt)->stack.thread)



/* ========================================================================= */
/* Kprocessor                                                                */
/* ========================================================================= */
typedef struct kaapi_perfproc_t {
  /* performance register used if library is configured with. Keep it for binary compatibility */
  kaapi_threadperf_counter_t perf_regs;
  kaapi_perf_counter_t*	     curr_perf_regs;   /* either perf_regs[0], either perf_regs[1] */
  int                        papi_event_set;
  unsigned int	             papi_event_count;
  kaapi_perf_counter_t       start_t[2];       /* [KAAPI_PERF_SCHEDULE_STATE]= T1 else = Tidle */
  uintptr_t                  serial;           /* serial number of generated steal */
  
  struct kaapi_event_buffer_t* eventbuffer;    /* event buffer */
} kaapi_perfproc_t;


/* ========================================================================= */
/* Kprocessor                                                                */
/* ========================================================================= */
/** \ingroup WS
    This data structure defines a work stealer processor kernel thread.
    A kprocessor is a container of tasks stored by stack.
    A kproc is part of a team (for non orphan processor). It may be bound
    to a ressource. 
    If hasnowork == 1, then it means that thread has no work and thief does
    not need to iterates over data structure to search for work.
A finest information would be that current context has no meaning because it is
used to receive task during work stealing request. TO TEST
*/
#define KAAPI_KPROC_SPECIFIC_SIZE 4
typedef struct kaapi_processor_t {
  kaapi_context_t*         context;                   /* current thread under execution */
  kaapi_team_t*            team;                      /* the team */
  kaapi_processor_id_t     kid;                       /* Kprocessor id in the context of a team */
  kaapi_numaid_t           numanode;                  /* numa node or -1 if not attached */
  kaapi_perfproc_t*        perf;                      /* !=0 if perf counters are enable */

  /* cache line shared by thieves & owner */
  kaapi_lock_t             lock                       /* used by aggr protocol to serialize thieves */
      __attribute__((aligned(KAAPI_CACHE_LINE)));
  uintptr_t volatile       hasnowork;                 /* if == 1, kproc has no work */
  /* queues of tasks to suspended thread
     task->body  : the entry point to restore the thread execution
     task->steal : signature int (fnc)(kaapi_task_t*, kaapi_thread_t* thief)
     task->arg   : the suspended thread
  */
  kaapi_lock_t             lock_lsuspend;
  kaapi_list_t             lsuspend;                  /* list of suspended context */

//  kaapi_atomic64_t         stealversion
//    __attribute__((aligned(KAAPI_CACHE_LINE)));


  /* private kproc fields */
  kaapi_places_part_t      place_part __attribute__((aligned(KAAPI_CACHE_LINE)));
  long                     place_off;   /* in place_part */
  kaapi_localitydomain_t*  lds[KAAPI_HWS_LEVELID_MAX]; /* hierarchy etc */
  kaapi_nodeinfo_t*        nodeinfos[KAAPI_HWS_LEVELID_MAX]; /* hierarchy etc */

  uint32_t                 seed;                      /* for the kproc own random generator */
  void*                    emitsteal_ctxt;            /* specific to the WS algorithm */
  uintptr_t                fnc_selecarg[4];           /* arguments for select victim function */
  unsigned int             proc_type;                 /* processor type */

  void*                    data_specific[KAAPI_KPROC_SPECIFIC_SIZE]; /* thread data specific */
  size_t                   size_specific[KAAPI_KPROC_SPECIFIC_SIZE]; /* size of data specific */

  char*                    sp_data;
  char                     stack __attribute__((aligned(8))); /* begin of stack for small allocation*/  
} kaapi_processor_t;


/* ========================================================================= */
/* Interface                                                                 */ 
/* ========================================================================= */

/* Barrier are strongly correlated to team. The number of threads in a barrier is
   the number of threads of the team.
*/
extern void kaapi_barrier_init (struct kaapi_barrier_t *barrier, int count);

/*
*/
extern void kaapi_barrier_destroy (struct kaapi_barrier_t *barrier);


/** Initialize a new team:
    - internal data structure is allocate on the kprocessor stack thread
    Return 0 in case of failure.

    * if kpl != 0 then procbind must be equal to default value.
    * if procbind == default value and kpl is not null, then the threads are created
    following information in kpl.
    * if procbind == default value and kpl is not null then the implementation
    will consider procbind as either master, close or spread

    EINVAL: if kpl ==0 and procbind != komp_proc_bind_default

*/
#define KAAPI_TEAM_DEFAULT 0
#define KAAPI_TEAM_NOCACHE 0x1
#define KAAPI_TEAM_LDSET   0x2  /* use topology information */
extern  kaapi_team_t* kaapi_team_init(
  kaapi_processor_t*         kproc,
  int                        count,
  kaapi_procbind_t           procbind,
  const kaapi_places_part_t* places,
  int                        flag
);

/** Wait end of a team, do not destroy data structure.
    Signal all threads to terminate their work and wait until the thread view the signal.
    After the call kaapi_self_processor() and kaapi_self_thread() return the master processor
    and master thread that exist previously to the call to kaapi_team_create...
*/
extern kaapi_thread_t* kaapi_team_finalize( kaapi_team_t* team );


/** Terminate a team
    Return 0 in case of success, else return an error code:
    EBUSY: one of the thread has remaining work not finished.
*/
extern int kaapi_team_destroy( kaapi_team_t* team, int force );


/** Start a new team
    The current thread is the thread 0 in the created team.
    The function creates a team following informations passed in the call kaapi_team_init

    Return 0 in case of success, else return an error code:
*/
extern int kaapi_team_start( kaapi_team_t* team );

/** \ingroup WS
    Higher level context manipulation.
    Assign context onto the running processor proc.
    This function is machine dependent.
*/
static inline int kaapi_setcontext( kaapi_processor_t* kproc, kaapi_context_t* ctxt )
{
  if (ctxt !=0)
    ctxt->proc    = kproc;
  kproc->context   = ctxt;
#if defined(KAAPI_HAVE_COMPILER_TLS_SUPPORT)
  kaapi_current_thread_key = &ctxt->stack.thread;
#endif
  return 0;
}
/*@}*/


/** \ingroup WS
    Alloc same data in the same memory region than the kproc data structure itself
*/
static inline void* kaapi_processor_alloc_data( kaapi_processor_t* kproc, size_t size )
{
  char* retval;
  size_t align =16;
  const uintptr_t mask = align - (uintptr_t)1;
  retval = (char*)((uintptr_t)(kproc->sp_data + mask) & ~mask);
  if ( ((retval + size) - (char*)kproc) >= KAAPI_KPROCESSOR_SIZE )
    return 0;
  kproc->sp_data = retval+size;
  return retval;
}


/** Return the kids of neighbor processor (in sens of a topology)
    Closed to kproc.
    Return value == the number of neghbors or -1 in case of failure.
*/
extern int kaapi_processor_getpu( kaapi_team_t* team, kaapi_processor_t* kproc,  int n, int* kids );


extern kaapi_stack_bloc_t* kaapi_processor_alloc_stackbloc(
    kaapi_processor_t* kproc
);

extern void kaapi_processor_dealloc_stackbloc(
    kaapi_processor_t* kproc, 
    kaapi_stack_bloc_t* bloc
);

/** Wait until no more theft on the kprocessor
*/
extern void kaapi_synchronize_steal( kaapi_processor_t* );

/** Wait until no more theft on the context
*/
extern void kaapi_synchronize_steal_context( kaapi_context_t* );


#if defined(KAAPI_HAVE_COMPILER_TLS_SUPPORT)
extern __thread kaapi_processor_t*      kaapi_current_processor_key;

/* */
static inline kaapi_processor_t* kaapi_self_processor(void)
{ return kaapi_current_processor_key; }

static inline kaapi_context_t* kaapi_self_context(void)
{ return kaapi_current_processor_key->context;}

static inline void kaapi_set_self_processor( kaapi_processor_t* kproc )
{
  kaapi_current_processor_key = kproc;
  if (kproc ==0)
    kaapi_current_thread_key    = 0;
  else
    kaapi_current_thread_key    = &kproc->context->stack.thread;
}

#else
extern pthread_key_t kaapi_current_processor_key;

static inline kaapi_processor_t* kaapi_self_processor(void)
{ return ((kaapi_processor_t*)pthread_getspecific( kaapi_current_processor_key )); }

static inline kaapi_context_t* kaapi_self_context(void)
{ return kaapi_self_processor()->context;}

static inline void kaapi_set_self_processor( kaapi_processor_t* kproc )
{
  kaapi_assert(0 == pthread_setspecific(kaapi_current_processor_key, kproc));
}
#endif


/*
*/
extern struct kaapi_processor_t* kaapi_create_orphan_kprocessor(void);

/*
*/
extern void kaapi_destroy_orphan_kprocessor(void);


/* Default entry point when CPU thread is created 
*/
extern void* kaapi_default_sched_run_processor( void* arg );


/** \ingroup TASK
    The function kaapi_context_alloc() allocates in the heap a context with a stack containing 
    at bytes for tasks and bytes for data.
    If successful, the kaapi_context_alloc() function will return a pointer to a kaapi_context_t.  
    Otherwise, an error number will be returned to indicate the error.
    This function is machine dependent.
    \param kproc IN/OUT the kprocessor that make allocation
    \retval pointer to the context
    \retval 0 if allocation failed
*/
extern kaapi_context_t* kaapi_context_alloc( struct kaapi_processor_t* kproc );

/** \ingroup TASK
    The function kaapi_context_destroy() free the ressource allocated with
    kaapi_context_alloc.
    If successful, the kaapi_context_destroy() function will return zero.
    Otherwise, an error number will be returned to indicate the error.
    This function is machine dependent.
    \param kproc IN/OUT the kprocessor that make allocation
    \param ctxt INOUT a pointer to the kaapi_context_t to allocate.
    \retval EINVAL invalid argument: bad stack pointer.
*/
extern int kaapi_context_destroy(
    kaapi_processor_t* kproc, 
    kaapi_context_t* ctxt 
);

/** \ingroup TASK
    The function kaapi_context_free() free the context successfuly allocated with 
    kaapi_context_alloc.
    If successful, the kaapi_context_free() function will return zero.  
    Otherwise, an error number will be returned to indicate the error.
    This function is machine dependent.
    \param kproc IN/OUT the kprocessor that make allocation
    \param ctxt INOUT a pointer to the kaapi_context_t to allocate.
    \retval EINVAL invalid argument: bad stack pointer.
*/
extern int kaapi_context_free( struct kaapi_processor_t* kproc, kaapi_context_t* ctxt );


/** \ingroup TASK
*/
extern int kaapi_setcontext( struct kaapi_processor_t* proc, kaapi_context_t* ctxt );


/** \ingroup TASK
    The function kaapi_context_clear re-initialize the context data structure
    If successful, the kaapi_context_clear() function returns 0.  
*/
extern int kaapi_context_clear( kaapi_context_t* ctxt );


/* ========================================================================== */
/** Termination dectecting barrier
*/
/**
*/
#define kaapi_barrier_td_init( kpb, value ) \
  KAAPI_ATOMIC_WRITE( kpb, value )

/**
*/
#define kaapi_barrier_td_destroy( kpb )
  
/**
*/
#define kaapi_barrier_td_setactive( kpb, b ) \
  if (b) { KAAPI_ATOMIC_INCR( kpb ); } \
  else KAAPI_ATOMIC_DECR( kpb )

/**
*/
#define kaapi_barrier_td_isterminated( kpb ) \
  (KAAPI_ATOMIC_READ(kpb ) == 0)

/**
*/
#define kaapi_barrier_td_waitterminated( kpb ) \
  while (KAAPI_ATOMIC_READ(kpb ) != 0)



/* 
*/
extern uint64_t kaapi_perf_thread_delayinstate(kaapi_processor_t* kproc);


/*
*/
static inline unsigned int kaapi_processor_get_type(const kaapi_processor_t* kproc)
{
  return kproc->proc_type;
}


/* lfree list routines
 */
static inline void kaapi_lfree_clear(kaapi_lfree_t* list)
{
  list->_sizelfree = 0;
  kaapi_list_init(&list->list);
}

static inline int kaapi_lfree_isempty( const kaapi_lfree_t* list)
{
  return list->_sizelfree == 0;
}

/* return 1 iff the ctxt is pushed into free list else return 0
*/
static inline int kaapi_lfree_push(
  kaapi_lfree_t* list, 
  struct kaapi_context_t* ctxt
)
{
#  define KAAPI_MAXFREECTXT 64
  if (list->_sizelfree >= KAAPI_MAXFREECTXT)
    return 0;

  kaapi_list_push_head(&list->list, &ctxt->stask);
  ++list->_sizelfree;

  return 1;
}


/* Re-design interface between context_free / context_alloc and lfree_push and lfree_pop
*/
static inline kaapi_context_t* kaapi_lfree_pop(kaapi_lfree_t* list)
{
  if (kaapi_list_empty(&list->list)) return 0;
  kaapi_task_t* stask = kaapi_list_pop_head(&list->list);
  kaapi_context_t* context = (kaapi_context_t*)stask->arg;
  --list->_sizelfree;
  return context;
}


/** Return true if the kprocess has no work in hierarchy queue
*/
static inline int kaapi_processor_has_nowork_hierarchy( const kaapi_processor_t* kproc )
{
  return 1;
}

/** Return true if the kprocess has no work == is begin stealing and suspend list is empty
*/
static inline int kaapi_processor_has_nowork( const kaapi_processor_t* kproc )
{
  return (kproc->hasnowork ==1);
}

/* */
extern void kaapi_mt_perf_init(void);
/* */
extern void kaapi_mt_perf_fini(void);
/* */
extern void kaapi_mt_perf_thread_init ( kaapi_perfproc_t* kproc, int isuser );
/* */
extern void kaapi_mt_perf_thread_fini ( kaapi_perfproc_t* kproc );
/* */
extern void kaapi_mt_perf_thread_start ( kaapi_perfproc_t* kproc );
/* */
extern void kaapi_mt_perf_thread_stop ( kaapi_perfproc_t* kproc );
/* */
extern void kaapi_mt_perf_thread_stopswapstart( kaapi_perfproc_t* kproc, int isuser );
/* must be called by the running thread */
extern void kaapi_mt_perf_thread_checkpoint( kaapi_perfproc_t* kproc );
/* */
extern uint64_t kaapi_mt_perf_thread_delayinstate(kaapi_perfproc_t* kproc);
/* */
extern size_t kaapi_mt_perf_counter_num(void);
/* */
extern const char* kaapi_mt_perf_id_to_name(kaapi_perf_id_t id);

#endif /* _KAAPI_MT_MACHINE_H */
