/*
 ** kaapi_task_checkdenpendencies.c
 ** xkaapi
 ** 
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 ** thierry.gautier@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"
#include <stdio.h>
#include <inttypes.h>

static int kaapi_tasklist_critical_path( kaapi_stack_t* stack, kaapi_frame_t* frame, kaapi_frame_wrdlist_t* rdlist );

/*
*/
int kaapi_frame_computereadylist_current(void)
{
  kaapi_context_t* ctxt = kaapi_self_context();
  kaapi_stack_t* stack  = &ctxt->stack;
  kaapi_frame_t* frame  = stack->unlink;
  return kaapi_frame_computereadylist(stack, frame);
}

/*
*/
int kaapi_frame_computereadylist(
  kaapi_stack_t* stack,
  kaapi_frame_t* frame
)
{
  int err = 0;
  kaapi_context_t* context = kaapi_stack2context(stack);
  kaapi_frame_wrdlist_t* rdlist = 0;
  kaapi_frame_dfgctxt_t*   dfgctxt = 0;
  kaapi_stack_iterator_t iter;


  rdlist = (kaapi_frame_wrdlist_t*)malloc( sizeof(kaapi_frame_wrdlist_t) + sizeof(kaapi_frame_dfgctxt_t) );
  kaapi_frame_rdlist_init( rdlist, (kaapi_frame_dfgctxt_t*)(rdlist+1));
  dfgctxt = rdlist->ctxt;

  err = kaapi_stack_iterator_init(&iter, stack, frame);
  if (err) goto return_retval;

  kaapi_task_t*              pc;
  kaapi_task_body_t          task_body;

#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_perf_counter_t    t1,t0;
  if (context->proc->perf !=0)
    t0 = kaapi_get_elapsedns();
#endif

  kaapi_assert_debug((frame->flag & KAAPI_FRAME_FLAG_DFG_OK) ==0);
  frame->flag |= KAAPI_FRAME_FLAG_NOTSTEAL;
  kaapi_writemem_barrier();

  while (!kaapi_stack_iterator_empty(&iter))
  {
    /* pass all task fully executed */
    pc = kaapi_stack_iterator_get( &iter );
    task_body = pc->body;
    if ((task_body == kaapi_nop_body)
     || (task_body == kaapi_taskmain_body)
    )
    {
      kaapi_stack_iterator_next( &iter );
      continue; 
    }
    kaapi_assert (pc->state.steal == pc->body );

    /* cannot assume pc->fmt !=0 : task->fmt not initialized if task->flag & DFG ==0 */
    const kaapi_format_t* task_fmt = pc->fmt = kaapi_format_resolvebybody(pc->body);
    kaapi_assert_debug((task_fmt !=0) || (pc->u.s.flag & KAAPI_TASK_FLAG_INDPENDENT));

    unsigned int war_param = 0;
    unsigned int cw_param  =0;
    kaapi_task_computeready(
              rdlist,
              pc,
              task_fmt,
              &war_param, 
              &cw_param,
              &dfgctxt->access_to_gd
    );
    kaapi_stack_iterator_next( &iter );
    ++rdlist->cnt_task;
  }

#undef KAAPI_DEBUG_INST
#define  KAAPI_DEBUG_INST(x) x

  /* Here compute the apriori minimal date of execution */
  if (kaapi_default_param.ctpriority)
  {
    KAAPI_DEBUG_INST(double t0 =) kaapi_get_elapsedtime();
    kaapi_tasklist_critical_path( &context->stack, frame, rdlist );
    KAAPI_DEBUG_INST(double t1 =) kaapi_get_elapsedtime();
    KAAPI_DEBUG_INST(printf("#tasks       = %"PRIu64"\n", rdlist->cnt_task);)
    KAAPI_DEBUG_INST(printf("Criticalpath = %"PRIu64"\n computed in :%f s\n", rdlist->t_infinity, t1-t0);)
#if 0
{
    static uint32_t counter = 0;
    char filename[128]; 
    if (getenv("USER") !=0)
      sprintf(filename,"/tmp/graph_ct.%s.%i.dot", getenv("USER"), counter++ );
    else
      sprintf(filename,"/tmp/graph_ct.%i.dot", counter++);
    FILE* filedot = fopen(filename, "w");
    kaapi_frame_print_dot( filedot, stack, frame);

    if (getenv("USER") !=0)
      sprintf(filename,"/tmp/graph_ct2.%s.%i.dot", getenv("USER"), counter++ );
    else
      sprintf(filename,"/tmp/graph_ct2.%i.dot", counter++);
    filedot = fopen(filename, "w");
    kaapi_frame_tasklist_print( filedot, rdlist);
    fclose(filedot);
}
#endif
  }
//printf("%i: Compute ready, #tasks %llu\n", kaapi_self_kid(), rdlist->cnt_task);
//WHY ?  frame->flag &= ~KAAPI_FRAME_FLAG_MASK_STATE;
  frame->rdlist = rdlist;
  kaapi_writemem_barrier();
  //frame->flag &= ~KAAPI_FRAME_FLAG_NOTSTEAL;
  frame->flag |= KAAPI_FRAME_FLAG_DFG_OK;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (context->proc->perf !=0)
  {
    t1 = kaapi_get_elapsedns();
    KAAPI_PERF_KPROC_REG_SYS(context->proc, KAAPI_PERF_ID_TASKLISTCALC) += t1-t0;
  }
#endif

  err= kaapi_hashmap_destroy( &dfgctxt->access_to_gd );

return_retval:
  return err;
}


/*
*/
static void kaapi_compute_cp( kaapi_task_t* task )
{
  uint64_t      maxdatesucc = 0;
  kaapi_access_mode_t next_mode_param;

  kaapi_assert_debug( task->u.s.flag & KAAPI_TASK_FLAG_DFGOK );
  const kaapi_format_t* fmt = task->fmt; 
  kaapi_assert_debug( fmt == kaapi_format_resolvebybody(task->body));
  unsigned int i, count_params = kaapi_format_get_count_params( fmt, task->arg );

  for (i=0; i<count_params; ++i)
  {
    kaapi_access_mode_t mode       = kaapi_format_get_mode_param(fmt, i, task->arg);
    kaapi_access_mode_t mode_param = KAAPI_ACCESS_GET_MODE( mode );
    if (mode_param & KAAPI_ACCESS_MODE_V)
      continue;

    /* look next successor */
    kaapi_access_t* access = kaapi_format_get_access_param(fmt, i, task->arg);
    access = access->next;
    if (access !=0)
    {
      uint64_t  datesucc = 0;
      kaapi_task_t* next_task = access->task;
      next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
      if (!KAAPI_ACCESS_IS_CONCURRENT( mode_param, next_mode_param))
      {
        datesucc = 1+next_task->u.s.date;
        //printf("Task :%p -> %p, access:%p, datesucc: %llu\n", (void*)task, (void*)next_task, access->data, datesucc );
      }
      else /* search next non concurrent */
      {
        while ((access !=0) && KAAPI_ACCESS_IS_CONCURRENT( mode_param, next_mode_param))
        {
          access = access->next;
          if (access)
          {
            next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
            next_task = access->task;
          }
        }
        if (access ==0)
          datesucc = 0;
        else
          datesucc = 1+next_task->u.s.date;
        //printf("Task :%p -> %p, access:%p, datesucc: %llu\n", (void*)task, (void*)next_task, (access == 0 ? 0 : access->data), datesucc );
      }

      if (datesucc > maxdatesucc)
        maxdatesucc = datesucc;
    }
  }
  kaapi_assert( maxdatesucc < UINT32_MAX);
  task->u.s.date = (uint32_t)maxdatesucc;
}


/* compute priority
*/
static void kaapi_compute_priority(
  kaapi_task_t* task,
  int version
)
{
  uint64_t      maxdatesucc = 0;
  kaapi_task_t* task_maxdatesucc = 0;
  kaapi_access_mode_t next_mode_param;
  kaapi_access_mode_t first_next_mode_param;

  /* if already visited, return */
  if (task->u.s.mark == version)
    return;
  kaapi_assert_debug( task->u.s.mark <= version );
  kaapi_assert_debug( task->u.s.flag & KAAPI_TASK_FLAG_DFGOK );

  const kaapi_format_t* fmt = kaapi_format_resolvebybody(task->body);
  unsigned int i, count_params = kaapi_format_get_count_params( fmt, task->arg );

  /* compute the max */
  for (i=0; i<count_params; ++i)
  {
    kaapi_access_mode_t mode       = kaapi_format_get_mode_param(fmt, i, task->arg);
    kaapi_access_mode_t mode_param = KAAPI_ACCESS_GET_MODE( mode );
    if (mode_param & KAAPI_ACCESS_MODE_V)
      continue;

    kaapi_access_t* access = kaapi_format_get_access_param(fmt, i, task->arg);

    /* iterate over all successors */
    access = access->next;
    if (access !=0)
    {
      uint64_t  datesucc = 0;
      kaapi_task_t* next_task = 0;
      next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
      if (!KAAPI_ACCESS_IS_CONCURRENT( mode_param, next_mode_param))
        next_task = access->task;
      else /* search next non concurrent */
      {
        while ((access !=0) && KAAPI_ACCESS_IS_CONCURRENT( mode_param, next_mode_param))
        {
          access = access->next;
          if (access)
          {
            next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
          }
        }
        if (access !=0)
          next_task = access->task;
      }

      datesucc = (next_task == 0 ? 0 : next_task->u.s.date);
      if (datesucc > maxdatesucc)
      {
        maxdatesucc = datesucc;
        task_maxdatesucc = next_task;
      }
    }
  }

  task->u.s.mark = version;
  task->u.s.flag |= KAAPI_TASK_FLAG_PRIORITY;
  task->u.s.priority = 1;

#if 0
  if (task_maxdatesucc)
    kaapi_compute_priority(task_maxdatesucc, version);
#else
  /* set the priority to all the max */
  for (i=0; i<count_params; ++i)
  {
    kaapi_access_mode_t mode       = kaapi_format_get_mode_param(fmt, i, task->arg);
    kaapi_access_mode_t mode_param = KAAPI_ACCESS_GET_MODE( mode );
    if (mode_param & KAAPI_ACCESS_MODE_V)
      continue;

    kaapi_access_t* access = kaapi_format_get_access_param(fmt, i, task->arg);

    /* iterate over all successors */
    access = access->next;
    if (access !=0)
    {
      uint64_t  datesucc = 0;
      kaapi_task_t* next_task = 0;
      next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
      if (!KAAPI_ACCESS_IS_CONCURRENT( mode_param, next_mode_param))
      {
#define SEUIL 1
        next_task = access->task;
        datesucc = (next_task == 0 ? 0 : next_task->u.s.date);
        if ((datesucc >= maxdatesucc-SEUIL) && (next_task !=0))
          kaapi_compute_priority(next_task, version);
        access = access->next;
        first_next_mode_param = next_mode_param;
        while ((access !=0) && KAAPI_ACCESS_IS_CONCURRENT( first_next_mode_param, access->mode))
        {
          next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
          next_task = access->task;
          datesucc = (next_task == 0 ? 0 : next_task->u.s.date);
          if ((datesucc >= maxdatesucc-SEUIL) && (next_task !=0))
            kaapi_compute_priority(next_task, version);
          access = access->next;
        }
      }
    }
  }
#endif
}


/* compute the longest path of each task
*/
static int kaapi_tasklist_critical_path( kaapi_stack_t* stack, kaapi_frame_t* frame, kaapi_frame_wrdlist_t* rdlist )
{
  if (kaapi_list_empty(&rdlist->rd))
    return EINVAL;

  uint64_t maxdate = 0;
  kaapi_task_t* task_maxdate = 0;

  kaapi_stack_iterator_t iter;
  kaapi_stack_reverse_iterator_init(&iter, stack, frame);

  while ( !kaapi_stack_iterator_empty(&iter))
  {
    kaapi_task_t* task = kaapi_stack_iterator_get(&iter);
    kaapi_compute_cp( task );
    if (task->u.s.date > maxdate)
    {
      maxdate = task->u.s.date;
      task_maxdate = task;
    }
    kaapi_stack_reverse_iterator_next(&iter);
  }

  rdlist->t_infinity = maxdate;

  if (task_maxdate)
    kaapi_compute_priority(task_maxdate, 1);

  kaapi_stack_reverse_iterator_destroy(&iter);
  return 0;
}
