/*
** xkaapi
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com 
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include "kaapi_impl.h"

/* This is the default, no aggregation steal protocol
*/


/* structure to retreive request object */
typedef struct kaapi_listrequest_t {
  kaapi_request_t        req;
} kaapi_listrequest_t;


/*
*/
static int kaapi_request_replytask_noaggr
( 
  kaapi_request_t*        request,
  int                     value
)
{
#if defined(KAAPI_DEBUG)
  int oldstatus = KAAPI_ATOMIC_READ(&request->status);
  kaapi_assert_debug( oldstatus == KAAPI_REQUEST_S_POSTED );
#endif
  if (value == KAAPI_REQUEST_S_OK)
    KAAPI_ATOMIC_WRITE(&request->status, KAAPI_REQUEST_S_OK);
  else
    KAAPI_ATOMIC_WRITE(&request->status, KAAPI_REQUEST_S_NOK);

  return 0;
}


/* The most important structure to iterate over the list of requests
   - only serve at most numserve request before stopping aggregation
*/
typedef struct kaapi_listrequest_iterator_noaggr_t {
  struct kaapi_listrequest_iterator_t lri;
  kaapi_request_t*                    curr;
} kaapi_listrequest_iterator_noaggr_t;


/* no concurrency here: always called before starting threads
*/
int kaapi_sched_noaggr_emitsteal_init(kaapi_processor_t* kproc)
{
  kaapi_listrequest_t* lr = kaapi_processor_alloc_data(kproc, sizeof(kaapi_listrequest_t));
  lr->req.ident    = kproc->kid;
  kproc->emitsteal_ctxt = lr;
  return 0;
}

/* no concurrency here: always called before starting threads 
*/
int kaapi_sched_noaggr_emitsteal_dstor(kaapi_processor_t* kproc)
{
  return 0;
}


/* return !=0 iff the range is empty
*/
static int kaapi_listrequest_noaggr_iterator_empty(struct kaapi_listrequest_iterator_t* lri)
{
  kaapi_listrequest_iterator_noaggr_t* lrrange = (kaapi_listrequest_iterator_noaggr_t*)lri;
  return lrrange->curr == 0;
}


/* get the first request of the range. range iterator should have been initialized by kaapi_listrequest_iterator_init 
*/
static kaapi_request_t* kaapi_listrequest_noaggr_iterator_get(
  struct kaapi_listrequest_iterator_t* lri
)
{ 
  kaapi_listrequest_iterator_noaggr_t* lrrange = (kaapi_listrequest_iterator_noaggr_t*)lri;
  return lrrange->curr;
}


/* return the next entry in the request. return 0 if the range is empty.
*/
static kaapi_request_t* kaapi_listrequest_noaggr_iterator_next(
  struct kaapi_listrequest_iterator_t* lri
)
{
  kaapi_listrequest_iterator_noaggr_t* lrrange = (kaapi_listrequest_iterator_noaggr_t*)lri;
  lrrange->curr = 0;
  return 0;
}


/*
*/
static int kaapi_listrequest_noaggr_iterator_count(
  struct kaapi_listrequest_iterator_t* lri
)
{
  kaapi_listrequest_iterator_noaggr_t* lrrange = (kaapi_listrequest_iterator_noaggr_t*)lri;
  if (lrrange->curr ==0) return 0;
  return 1;
}


/*
*/
static inline void kaapi_listrequest_iterator_init(
          kaapi_listrequest_t* lr,
          kaapi_listrequest_iterator_noaggr_t* lri
)
{
  lri->curr = &lr->req;
  lri->lri.lri_empty = kaapi_listrequest_noaggr_iterator_empty;
  lri->lri.lri_get   = kaapi_listrequest_noaggr_iterator_get;
  lri->lri.lri_next  = kaapi_listrequest_noaggr_iterator_next;
  lri->lri.lri_count = kaapi_listrequest_noaggr_iterator_count;
}


/*
*/
static kaapi_request_t* kaapi_request_post(
  kaapi_listrequest_t*   own_lrequests,
  kaapi_processor_id_t   selfid,
  const kaapi_thread_t*  thread
)
{
  kaapi_request_t* self = &own_lrequests->req;
  self->reply  = kaapi_request_replytask_noaggr;
  self->ident  = (uintptr_t)selfid;
  self->thread = *thread;
  KAAPI_ATOMIC_WRITE(&self->status, KAAPI_REQUEST_S_POSTED);
  return self;
}



/*
*/
kaapi_request_status_t kaapi_sched_noaggr_emitsteal ( kaapi_processor_t* kproc )
{
  kaapi_team_t*                team;
  kaapi_victim_t               victim;
  kaapi_request_t*             self_request;
  int                          err;

#if defined(KAAPI_USE_PERFCOUNTER)
  uintptr_t serial;
  kaapi_perf_counter_t date0 = 123124;
  kaapi_perf_counter_t date1 = 123124;
#endif

  team = kproc->team;
  if (team->count <2) 
    return KAAPI_REQUEST_S_NOK;

redo_select:
  /* select the victim processor */
  err = (*team->fnc_select)( kproc, &victim, KAAPI_SELECT_VICTIM );
  if (unlikely(err !=0)) 
    goto redo_select;

  if (victim.kproc ==0) goto redo_select;
  kaapi_assert_debug( (victim.kproc->kid >=0) && (victim.kproc->kid < (uintptr_t)team->count));

  /* quick test to detect if thread has no work */
  if (kaapi_processor_has_nowork(victim.kproc))
  {
    (*team->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );
    return KAAPI_REQUEST_S_NOK;
  }

  /* No aggregation: serialize thieves on the victim lock
  */
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    date0 = kaapi_get_elapsedns();
#endif

  /* lock victim processor */
  kaapi_atomic_lock( &victim.kproc->lock );

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TSYNC) += date1-date0;
    ++kproc->perf->serial;
  }
#endif

  /* Post a request to the victim processor
  */
  kaapi_listrequest_t* own = (kaapi_listrequest_t*)kproc->emitsteal_ctxt;
  kaapi_listrequest_iterator_noaggr_t lri;
  self_request = kaapi_request_post(
    own,
    kproc->kid,
    kaapi_context2thread(kproc->context)
  );

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perfproc_t* perf = kproc->perf;
    self_request->victim = (uintptr_t)victim.kproc->kid;
    serial = ++kproc->perf->serial;
    self_request->serial = serial;
    KAAPI_EVENT_PUSH2(perf, 0, KAAPI_EVT_STEAL_OP,
        (uintptr_t)victim.kproc->kid, 
        self_request->serial
    );
    ++KAAPI_PERF_REG(kproc->perf, KAAPI_PERF_ID_STEALREQ);
    KAAPI_EVENT_PUSH2(kproc->perf, 0, KAAPI_EVT_REQUESTS_BEG,
                      (uintptr_t)victim.kproc->kid, serial );
  }
#endif

  /* Dummy listrequest with my self_request inside */
  kaapi_listrequest_iterator_init(
          own,
          &lri
  );

  if (!kaapi_processor_has_nowork(victim.kproc))
  {
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
      ++KAAPI_PERF_REG(kproc->perf, KAAPI_PERF_ID_STEALOP);
#endif
    kaapi_sched_stealprocessor( victim.kproc, (struct kaapi_listrequest_iterator_t*)&lri );
  }

  /* release lock */
  kaapi_atomic_unlock( &victim.kproc->lock );
  
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    KAAPI_EVENT_PUSH2(kproc->perf, 0, KAAPI_EVT_REQUESTS_END,
              (uintptr_t)victim.kproc->kid, serial
    );
#endif

  if (kaapi_request_test( self_request ))
    return kaapi_sched_emitsteal_return( kproc, &victim, self_request );
  
  return KAAPI_REQUEST_S_NOK;
}
