/*
 ** xkaapi
 ** 
 ** Created on Tue Mar 31 15:21:00 2009
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"
//#include <unistd.h> //usleep
#include <time.h> // nanosleep

#if defined(KAAPI_DEBUG)
#include <stdio.h>
#endif

/* arg = argstruct for the steal state method */
#if 1 /* for High perf in dpotrf */
#define KAAPI_STEAL_FAILED_INST( arg )\
  { \
     ++arg->count; \
     kaapi_slowdown_cpu(); \
  }
#else
#define KAAPI_STEAL_FAILED_INST( arg )\
   ++arg->count; \
   if (arg->count > 8) \
   { \
     if (arg->delay ==0) arg->delay = 1;\
     struct timespec ts = { 0, arg->delay }; \
     nanosleep( &ts, 0 ); \
     arg->delay = 2*arg->delay; \
     if (arg->delay > 64) arg->delay = 1; \
   }
#endif

#define KAAPI_STEAL_SUCCESS_INST( arg )\
    arg->count = 0;\
    arg->delay = 1;\

/* linear backoff */
//   usleep(5 * (arg->count % 10) )


/* fnc_selecarg[Ø] == lastsuccess_kid +1 */
typedef struct kaapi_victim_rand_arg_t {
  int init;
  int lastvictimid;
  int count;
  int delay;
} kaapi_victim_rand_arg_t;

/** Do rand selection 
 */
int kaapi_sched_select_victim_rand
(
    kaapi_processor_t* kproc, 
    kaapi_victim_t* victim, 
    kaapi_selecvictim_flag_t flag 
)
{
  kaapi_team_t* team = kproc->team;
  kaapi_victim_rand_arg_t* arg = (kaapi_victim_rand_arg_t*)kproc->fnc_selecarg;
   
  switch (flag)
  {
    case KAAPI_SELECT_VICTIM:
    {
      int nbproc, victimid;
      kaapi_processor_t* victim_proc = 0;
      {
        nbproc = team->count;
        if (nbproc <=1) 
          return EINVAL;
        victimid = rand_r( (unsigned int*)&kproc->seed /*fnc_selecarg*/ ) % nbproc;
      }
      
      /* Get the k-processor */    
      victim_proc = team->all_kprocessors[ victimid ];

      victim->kid   = victimid;
      victim->kproc = victim_proc;

      break;
    }
      
    case KAAPI_STEAL_SUCCESS:
    {
      KAAPI_STEAL_SUCCESS_INST(arg);
      break;
    }
    case KAAPI_STEAL_FAILED:
    {
      KAAPI_STEAL_FAILED_INST(arg);
      break;
    }

    default:
    {
      break;
    }
  }
  
  return 0;
}



/** Do hierarchical selection up to a threshold level
*/
typedef struct kaapi_victim_hws_arg_t {
  int level;
  int count;
  int delay;
} kaapi_victim_hws_arg_t;


int kaapi_sched_select_victim_hierarchy(
    kaapi_processor_t*       kproc, 
    kaapi_victim_t*          victim,
    kaapi_selecvictim_flag_t flag 
)
{
  kaapi_victim_hws_arg_t* arg = (kaapi_victim_hws_arg_t*)kproc->fnc_selecarg;
#if defined(__MIC__)
  int cpuid = kaapi_processor_getcpuid(kproc);
  int coreid = cpuid -1;
  if (coreid <0) coreid += kaapi_default_param.syscpucount;
  coreid /= 4; /* for series 5100, 3100, 7100 */
#endif

  switch (flag)
  {
    case KAAPI_SELECT_VICTIM:
    {
      kaapi_localitydomain_t* ld = kproc->lds[arg->level];
      if (arg->count >= 2 * ld->count)
      {
        arg->count = 0;
        do {
          ++arg->level;
          if (arg->level >= KAAPI_HWS_LEVELID_MAX) /* here threshold ? */
          {
            arg->level = 0;
            break;
          }
          ld = kproc->lds[arg->level];
        } while ((ld ==0) || (ld->count == 1));
        if (ld ==0) ld = kproc->lds[arg->level = 0];
      }
      int nbproc = ld->count;
      int victimid;

      victimid = rand_r( (unsigned int*)&kproc->seed ) % nbproc;

      /* Get the k-processor */    
      victim->kproc = ld->kids[victimid];
      if (victim->kproc ==0) return EINVAL;
      victim->kid   = victim->kproc->kid;
      return 0;
    } break;

    case KAAPI_STEAL_SUCCESS:
    {
      KAAPI_STEAL_SUCCESS_INST(arg);
    } break;

    case KAAPI_STEAL_FAILED:
    {
      KAAPI_STEAL_FAILED_INST(arg);
    } break;

    default:
      break;
  }

  return 0;
}

/** Do local (intra numa node) steal, then other numa node
*/
typedef struct kaapi_victim_numa_arg_t {
  int level;
  int count;
  int delay;
} kaapi_victim_numa_arg_t;


//#define UNIQUE

int kaapi_sched_select_victim_numa(
    kaapi_processor_t*       kproc, 
    kaapi_victim_t*          victim,
    kaapi_selecvictim_flag_t flag 
)
{
  kaapi_victim_numa_arg_t* arg = (kaapi_victim_numa_arg_t*)kproc->fnc_selecarg;
  kaapi_localitydomain_t* ld;

  switch (flag)
  {
    case KAAPI_SELECT_VICTIM:
    {
      if (arg->level ==0) 
        arg->level = KAAPI_HWS_LEVELID_NUMA;

      /* only one thread at once could emits steal request to inter-numa node */
      ld = kproc->lds[arg->level];
#ifdef UNIQUE
      if ((arg->level == KAAPI_HWS_LEVELID_MACHINE) && !kaapi_atomic_trylock(&ld->lock_up))
        arg->level = KAAPI_HWS_LEVELID_NUMA;
#endif
      ld = kproc->lds[arg->level];
      int nbproc = ld->count;
      int victimid = rand_r( (unsigned int*)&kproc->seed ) % nbproc;
      victim->kproc = ld->kids[victimid];
      if (victim->kproc ==0) return EINVAL;
      victim->kid   = victim->kproc->kid;
      return 0;
    } break;

    case KAAPI_STEAL_SUCCESS:
    {
#ifdef UNIQUE
      if (arg->level == KAAPI_HWS_LEVELID_MACHINE)
      {
        ld = kproc->lds[arg->level];
        kaapi_atomic_unlock(&ld->lock_up);
        arg->level = KAAPI_HWS_LEVELID_NUMA;
      }
#endif
      arg->level = KAAPI_HWS_LEVELID_NUMA;
      KAAPI_STEAL_SUCCESS_INST(arg);
    } break;

    case KAAPI_STEAL_FAILED:
    {
      ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA];
#ifdef UNIQUE
      if (arg->level == KAAPI_HWS_LEVELID_MACHINE)
        kaapi_atomic_unlock(&ld->lock_up);
#endif
#if 0
      KAAPI_STEAL_FAILED_INST(arg);
      if (arg->count >=  16*ld->count) 
      {
        if (arg->level == KAAPI_HWS_LEVELID_NUMA)
          arg->level = KAAPI_HWS_LEVELID_MACHINE;
      }
#endif
    } break;

    default:
      break;
  }

  return 0;
}
