/*
 ** xkaapi
 ** 
 ** Created on Tue Mar 31 15:21:00 2009
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"
#if defined(KAAPI_USE_NUMA)
#include <numa.h>
#include <numaif.h>
#endif


/** Local strategy first
*/
int kaapi_sched_select_queue_local(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t* task,
    kaapi_queue_t* local_queue
)
{
  return kaapi_queue_push_head( local_queue, task ); // impact here ?
}


/**
*/
int kaapi_sched_select_queue_attr(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t* task,
    kaapi_queue_t* local_queue
)
{
  if (task->u.s.site == 0)
    return kaapi_sched_select_queue_local( kproc, frame_rd, task, local_queue );
  else
  {
    int ldid = task->u.s.site -1;
    kaapi_team_t* team = kproc->team;
    if (ldid >= team->localitydomains[KAAPI_HWS_LEVELID_NUMA].count)
    {
#if defined(KAAPI_DEBUG)
      fprintf(stderr, "*** warning 'task attribut ld is out of bound: %i'\n", ldid);
#endif
      task->u.s.site =0;
      return kaapi_sched_select_queue_localnuma( kproc, frame_rd, task, local_queue);
    }

    kaapi_localitydomain_t* ld = team->localitydomains[KAAPI_HWS_LEVELID_NUMA].lds[ldid];
    if (ld ==0) /* no core attached to this ldid, use NUMA ld of the running kproc */
      ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA];

    int err;
    kaapi_atomic_lock( &ld->lock_r );
    if (task->u.s.flag & KAAPI_TASK_FLAG_LD_BOUND)
    {
      //printf("*** push: %p on ld private queue %i\n", (void*)task, ld->lid );
      err = kaapi_queue_push_head( &ld->private_queue, task );
    }
    else
      err = kaapi_queue_push_head( &ld->queue, task );
    kaapi_atomic_unlock( &ld->lock_r );
    if (err)
      return kaapi_sched_select_queue_local(kproc, frame_rd, task, local_queue);
  }
  return 0;
}


/**
*/
int kaapi_sched_select_queue_localnuma(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t* task,
    kaapi_queue_t* local_queue
)
{
  if (task->u.s.site !=0)
    return kaapi_sched_select_queue_attr(kproc, frame_rd, task, local_queue );

  if (task->u.s.flag & KAAPI_TASK_FLAG_PRIORITY)
    return kaapi_queue_push_head( local_queue, task );

  kaapi_localitydomain_t* ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA];
  int err;
  kaapi_atomic_lock( &ld->lock_r );
  if (task->u.s.flag & KAAPI_TASK_FLAG_LD_BOUND)
    err = kaapi_queue_push_head( &ld->private_queue, task );
  else
    err = kaapi_queue_push_head( &ld->queue, task );
  kaapi_atomic_unlock( &ld->lock_r );
  if (err)
    return kaapi_queue_push_tail( local_queue, task ); 
  return 0;
}


/** Return the numa queue were a W param of the task is stored.
*/
int kaapi_sched_select_queue_Wnuma(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t* task,
    kaapi_queue_t* local_queue
)
{
  if (task->u.s.site !=0)
    return kaapi_sched_select_queue_attr(kproc, frame_rd, task, local_queue );
  if ((task->u.s.flag & KAAPI_TASK_FLAG_PRIORITY)) 
    return kaapi_queue_push_head( local_queue, task );

  const kaapi_format_t* fmt = task->fmt;
  kaapi_team_t* team = kproc->team;

#if 0
  size_t size;
  if ((size = _kaapi_task_getsize(task,fmt)) < 196608)
    return kaapi_sched_select_queue_localnuma(kproc, frame_rd, task, local_queue);
  const kaapi_format_t* fmt = task->fmt;
  kaapi_team_t* team = kproc->team;
#endif

  unsigned int i;
  unsigned int count_params = kaapi_format_get_count_params( fmt, task->arg );

  for (i=0; i<count_params; ++i)
  {
    kaapi_access_mode_t mode            = kaapi_format_get_mode_param(fmt, i, task->arg);
    if (!KAAPI_ACCESS_IS_WRITE(mode)) continue;

    kaapi_access_t* access = kaapi_format_get_access_param(fmt, i, task->arg);
    int numanode = kaapi_numa_getpage_id(access->data);
    if (numanode != -1)
    {
      /* find the numa queue for node */
      if (numanode < team->localitydomains[KAAPI_HWS_LEVELID_NUMA].count)
      {
        kaapi_localitydomain_t* ld = team->localitydomains[KAAPI_HWS_LEVELID_NUMA].lds[numanode];
        if (ld ==0) ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA];
        int err;

        kaapi_atomic_lock( &ld->lock_r );
        if (task->u.s.flag & KAAPI_TASK_FLAG_LD_BOUND)
          err = kaapi_queue_push_head( &ld->private_queue, task );
        else
          err = kaapi_queue_push_head( &ld->queue, task );
        kaapi_atomic_unlock( &ld->lock_r );

        if (err)
          return kaapi_sched_select_queue_local(kproc, frame_rd, task, local_queue);
        return 0;
      }
      else
        kaapi_abort(__LINE__, __FILE__, __FUNCTION__);
    }
  }
  return kaapi_sched_select_queue_localnuma(kproc, frame_rd, task, local_queue);
}

