/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"

/*
*/
int kaapi_request_replytask
( 
  kaapi_request_t*        request,
  kaapi_request_status_t  value
)
{
  return request->reply(request,value);
}


/*
*/
struct kaapi_request_t* kaapi_listrequest_iterator_get(
  struct kaapi_listrequest_iterator_t* lrrange
)
{
  return lrrange->lri_get(lrrange);
}


/*
*/
struct kaapi_request_t* kaapi_listrequest_iterator_next(
  struct kaapi_listrequest_iterator_t* lrrange 
)
{
  return lrrange->lri_next(lrrange);
}


/*
*/
int kaapi_listrequest_iterator_empty(
  struct kaapi_listrequest_iterator_t* lrrange
)
{
  return lrrange->lri_empty(lrrange);
}


/*
*/
int kaapi_listrequest_iterator_count(
  struct kaapi_listrequest_iterator_t* lrrange
)
{
  return lrrange->lri_count(lrrange);
}


/** \ingroup WS
    \retval 0 if no context could be wakeup
    \retval else a context to wakeup
    \TODO faire specs ici
*/
static kaapi_context_t* kaapi_sched_wakeup (
  kaapi_processor_t* kproc
);


/*
*/
int kaapi_sched_advance ( kaapi_processor_t* kproc )
{
#if defined(KAAPI_USE_NETWORK)
    kaapi_network_poll();
#endif
#if defined(KAAPI_USE_CUDA)
    if( kaapi_processor_get_type(kproc) == KAAPI_PROC_TYPE_CUDA ) 
      kaapi_cuda_proc_poll( kproc );
#endif

  return 0;
}



/* Suspend the current running kproc until the predicat fcondition(arg_fcondition) is true (!=0).
   The implementation relies on pushing in LIFO link list kaapi_queue_t a task that represent
   the suspension point. This task has fields:
    .body  = fcondition
    .steal = arg_fcondition
    .arg   = the current running thread
*/
int kaapi_sched_suspend ( 
    kaapi_processor_t* kproc, 
    int (*fcondition)(void* ), 
    void* arg_fcondition
)
{
  int err = 0;
  uintptr_t savednowork = kproc->hasnowork;
  kaapi_request_status_t ws_status;
  kaapi_context_t* thread;
  kaapi_context_t* context_input = 0;
  kaapi_context_t* tmp;
  kaapi_task_t task_suspended;
  kaapi_team_t* team;

  kaapi_assert_debug( kproc !=0 );
  kaapi_assert_debug( kproc->context !=0 );
  kaapi_assert_debug( kproc == kaapi_self_processor() );

  team = kproc->team;

  if ((fcondition !=0) && fcondition(arg_fcondition))
    return 0;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perfproc_t* perf = kproc->perf;
    kaapi_perf_thread_stopswapstart(perf, KAAPI_PERF_SCHEDULE_STATE );
    KAAPI_EVENT_PUSH0( perf, 0, KAAPI_EVT_SCHED_IDLE_BEG );
  }
#endif

  /* here is the reason of suspension */
  context_input = kproc->context;
  if (context_input !=0)
  {
    kaapi_assert_debug( kproc == context_input->proc);
    /* put context in the list of suspended contexts: no critical section with respect of thieves */
    kaapi_setcontext(kproc, 0);
    task_suspended.body  = (kaapi_task_body_t)fcondition;
    task_suspended.state.steal = (kaapi_task_body_t)arg_fcondition;
    task_suspended.arg   = context_input;
    task_suspended.next  = 0;
    task_suspended.prev  = 0;
    kaapi_atomic_lock(&kproc->lock_lsuspend);
    kaapi_list_push_head(&kproc->lsuspend, &task_suspended);
    kaapi_atomic_unlock(&kproc->lock_lsuspend);

    /* always allocate a thread before stealing */
    thread = kaapi_context_alloc(kproc);
    kaapi_assert_debug( thread != 0 );
    kaapi_setcontext(kproc, thread);
  }

  do
  {
    kaapi_sched_advance(kproc);

    /* wakeup a context: either a ready thread (first) or a suspended thread.
       Precise the suspended thread 'context_input' in order to wakeup it first and task_condition.
    */
    if (!kaapi_list_empty(&kproc->lsuspend))
    {
      thread = kaapi_sched_wakeup(kproc);
      if (thread !=0)
      {
        tmp = kproc->context;
        kaapi_setcontext( kproc, thread );
        kaapi_context_destroy( kproc, tmp );

        if ((fcondition !=0) && (thread == context_input))
        {
          err = 0;
          break;
        }
        
        goto redo_execute;
      }
    }

    /*local steal first ? */
    if ((team->fnc_localsteal !=0) && team->fnc_localsteal(kproc))
      goto start_execute;
    
    /* steal request */
    ws_status = team->emitsteal(kproc);
    if (ws_status != KAAPI_REQUEST_S_OK)
    {
      if ((ws_status == KAAPI_REQUEST_S_STOP) || team->isterm)
      {
        err = EINTR;
        break;
      }
      continue;
    }

start_execute:

redo_execute:
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_USER_STATE );
      KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_END );
    }
#endif
    if (savednowork == 1)
      kproc->hasnowork = 0;
    thread = kproc->context;
    err = kaapi_sched_sync( kaapi_context2thread(thread) );
    kaapi_assert( err ==0);
    kproc->hasnowork = savednowork;

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_SCHEDULE_STATE );
      KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_BEG );
    }
#endif
    /* synchronize on kproc in order to avoid stealer viewing the stack that will be reset */
    kproc->context = 0;
//    kaapi_synchronize_steal(kproc);
    kaapi_stack_reset(&thread->stack);
    kaapi_writemem_barrier();
    kproc->context = thread;
  } while (1);

  kproc->hasnowork = savednowork;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_USER_STATE );
    KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_END );
  }
#endif
  return err;
}


/*
*/
int kaapi_sched_idle (
    kaapi_processor_t* kproc,
    int (*fcondition)(void* ),
    void* arg_fcondition
)
{
  int err = 0;

  uintptr_t savehasnowork = kproc->hasnowork;
  kaapi_request_status_t  ws_status;
  kaapi_context_t* thread;
  kaapi_context_t* tmp;
  kaapi_team_t* team;

  kaapi_assert_debug( kproc !=0 );
  kaapi_assert_debug( kproc->context !=0 );
  kaapi_assert_debug( kproc == kaapi_self_processor() );

  team = kproc->team;

  /* ok now we can mark that the kproc has it has no work */
  kproc->hasnowork = 1;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perfproc_t* perf = kproc->perf;
    kaapi_perf_thread_stopswapstart(perf, KAAPI_PERF_SCHEDULE_STATE );
    KAAPI_EVENT_PUSH0( perf, 0, KAAPI_EVT_SCHED_IDLE_BEG );
  }
#endif

  do {
    kaapi_sched_advance(kproc);

    /* wakeup a context: either a ready thread (first) or a suspended thread.
       Precise the suspended thread 'context_input' in order to wakeup it first and task_condition.
    */
    if (!kaapi_list_empty(&kproc->lsuspend))
    {
      thread = kaapi_sched_wakeup(kproc);
      if (thread !=0)
      {
        tmp = kproc->context;
        kaapi_setcontext( kproc, thread );
        kaapi_context_destroy( kproc, tmp );
        goto redo_execute;
      }
    }

    /*local steal first ? */
    if ((team->fnc_localsteal !=0) && team->fnc_localsteal(kproc))
      goto start_execute;
    
    /* steal request */
    ws_status = team->emitsteal(kproc);
    if (ws_status != KAAPI_REQUEST_S_OK)
    {
      if ((ws_status == KAAPI_REQUEST_S_STOP) || team->isterm)
      {
        err = EINTR;
        break;
      }

      /* return if condition is true */
      if ((fcondition !=0) && (fcondition(arg_fcondition) !=0))
        break;

      //sched_yield();
      continue;
    }

start_execute:

redo_execute:
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_USER_STATE );
      KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_END );
    }
#endif
    kproc->hasnowork = 0;
    thread = kproc->context;
    err = kaapi_sched_sync( kaapi_context2thread(thread) );
    kaapi_assert( err ==0);
    kproc->hasnowork = 1;

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_SCHEDULE_STATE );
      KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_BEG );
    }
#endif
    /* synchronize on kproc in order to avoid stealer viewing the stack that will be reset */
    kproc->context = 0;
//    kaapi_synchronize_steal(kproc);
    kaapi_stack_reset(&thread->stack);
    kaapi_writemem_barrier();
    kproc->context = thread;
  } while (1);


  kproc->hasnowork = savehasnowork;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_USER_STATE );
    KAAPI_EVENT_PUSH0( kproc->perf, 0, KAAPI_EVT_SCHED_IDLE_END );
  }
#endif

  return err;
}



/* See comment before definition of kaapi_sched_suspend
*/
static kaapi_context_t* kaapi_sched_wakeup (
    kaapi_processor_t*      kproc
)
{
  kaapi_task_t* stask;
    
  /* Note on this line: in the current implementation, only the owner of the suspend queue
     is able to call kaapi_sched_wakeup.
  */
  kaapi_assert_debug( kproc->kid == kaapi_self_kid() );

  /* else...
     - here only do garbage, because in the current version the 
     thread that becomes ready is pushed into an other queue.
     Thus, all cell which are marked KAAPI_WSQUEUECELL_OUTLIST or
     KAAPI_WSQUEUECELL_STEALLIST are garbaged.
  */
  stask = kproc->lsuspend.head;
  if (stask ==0)
    return 0;
  kaapi_isready_fnc_t isready = (kaapi_isready_fnc_t)stask->body;
  void* arg_isready = (void*)stask->state.steal;

  if (isready(arg_isready))
  {
    kaapi_context_t* context = (kaapi_context_t*)stask->arg;
    kaapi_atomic_lock(&kproc->lock_lsuspend);
    kaapi_list_pop_head(&kproc->lsuspend);
    kaapi_atomic_unlock(&kproc->lock_lsuspend);
    return context;
  }

  return 0; 
}
